""" Module containing the main model class for the dowhy package.

"""
import logging
from itertools import combinations

from sympy import init_printing

import dowhy.causal_estimators as causal_estimators
import dowhy.causal_refuters as causal_refuters
import dowhy.graph_learners as graph_learners
import dowhy.utils.cli_helpers as cli
from dowhy.causal_estimator import CausalEstimate, estimate_effect
from dowhy.causal_graph import CausalGraph
from dowhy.causal_identifier import AutoIdentifier, BackdoorAdjustment, IDIdentifier
from dowhy.causal_identifier.identify_effect import EstimandType
from dowhy.causal_refuters.graph_refuter import GraphRefuter
from dowhy.utils.api import parse_state

init_printing()  # To display symbolic math symbols


class CausalModel:

    """Main class for storing the causal model state."""

    def __init__(
        self,
        data,
        treatment,
        outcome,
        graph=None,
        common_causes=None,
        instruments=None,
        effect_modifiers=None,
        estimand_type="nonparametric-ate",
        proceed_when_unidentifiable=False,
        missing_nodes_as_confounders=False,
        identify_vars=False,
        **kwargs,
    ):
        """Initialize data and create a causal graph instance.

        Assigns treatment and outcome variables.
        Also checks and finds the common causes and instruments for treatment
        and outcome.

        At least one of graph, common_causes or instruments must be provided. If
        none of these variables are provided, then learn_graph() can be used later.

        :param data: a pandas dataframe containing treatment, outcome and other
        variables.
        :param treatment: name of the treatment variable
        :param outcome: name of the outcome variable
        :param graph: path to DOT file containing a DAG or a string containing
        a DAG specification in DOT format
        :param common_causes: names of common causes of treatment and _outcome. Only used when graph is None.
        :param instruments: names of instrumental variables for the effect of
        treatment on outcome. Only used when graph is None.
        :param effect_modifiers: names of variables that can modify the treatment effect. If not provided, then the causal graph is used to find the effect modifiers. Estimators will return multiple different estimates based on each value of effect_modifiers.
        :param estimand_type: the type of estimand requested (currently only "nonparametric-ate" is supported). In the future, may support other specific parametric forms of identification.
        :param proceed_when_unidentifiable: does the identification proceed by ignoring potential unobserved confounders. Binary flag.
        :param missing_nodes_as_confounders: Binary flag indicating whether variables in the dataframe that are not included in the causal graph, should be  automatically included as confounder nodes.
        :param identify_vars: Variable deciding whether to compute common causes, instruments and effect modifiers while initializing the class. identify_vars should be set to False when user is providing common_causes, instruments or effect modifiers on their own(otherwise the identify_vars code can override the user provided values). Also it does not make sense if no graph is given.
        :returns: an instance of CausalModel class

        """
        self._data = data
        self._treatment = parse_state(treatment)
        self._outcome = parse_state(outcome)
        self._effect_modifiers = parse_state(effect_modifiers)
        self._estimand_type = estimand_type
        self._proceed_when_unidentifiable = proceed_when_unidentifiable
        self._missing_nodes_as_confounders = missing_nodes_as_confounders
        self.logger = logging.getLogger(__name__)
        self._estimator_cache = {}

        if graph is None:
            self.logger.warning("Causal Graph not provided. DoWhy will construct a graph based on data inputs.")
            self._common_causes = parse_state(common_causes)
            self._instruments = parse_state(instruments)
            if common_causes is not None and instruments is not None:
                self._graph = CausalGraph(
                    self._treatment,
                    self._outcome,
                    common_cause_names=self._common_causes,
                    instrument_names=self._instruments,
                    effect_modifier_names=self._effect_modifiers,
                    observed_node_names=self._data.columns.tolist(),
                )
            elif common_causes is not None:
                self._graph = CausalGraph(
                    self._treatment,
                    self._outcome,
                    common_cause_names=self._common_causes,
                    effect_modifier_names=self._effect_modifiers,
                    observed_node_names=self._data.columns.tolist(),
                )
            elif instruments is not None:
                self._graph = CausalGraph(
                    self._treatment,
                    self._outcome,
                    instrument_names=self._instruments,
                    effect_modifier_names=self._effect_modifiers,
                    observed_node_names=self._data.columns.tolist(),
                )
            else:
                self.logger.warning(
                    "Relevant variables to build causal graph not provided. You may want to use the learn_graph() function to construct the causal graph."
                )
                self._graph = CausalGraph(
                    self._treatment,
                    self._outcome,
                    effect_modifier_names=self._effect_modifiers,
                    observed_node_names=self._data.columns.tolist(),
                )
        else:
            self.init_graph(graph=graph, identify_vars=identify_vars)

        self._other_variables = kwargs
        self.summary()

    def init_graph(self, graph, identify_vars):
        """
        Initialize self._graph using graph provided by the user.

        """
        # Create causal graph object
        self._graph = CausalGraph(
            self._treatment,
            self._outcome,
            graph,
            effect_modifier_names=self._effect_modifiers,
            observed_node_names=self._data.columns.tolist(),
            missing_nodes_as_confounders=self._missing_nodes_as_confounders,
        )

        if identify_vars:
            self._common_causes = self._graph.get_common_causes(self._treatment, self._outcome)
            self._instruments = self._graph.get_instruments(self._treatment, self._outcome)
            # Sometimes, effect modifiers from the graph may not match those provided by the user.
            # (Because some effect modifiers may also be common causes)
            # In such cases, the user-provided modifiers are used.
            # If no effect modifiers are provided,  then the ones from the graph are used.
            if self._effect_modifiers is None or not self._effect_modifiers:
                self._effect_modifiers = self._graph.get_effect_modifiers(self._treatment, self._outcome)

    def get_common_causes(self):
        self._common_causes = self._graph.get_common_causes(self._treatment, self._outcome)
        return self._common_causes

    def get_instruments(self):
        self._instruments = self._graph.get_instruments(self._treatment, self._outcome)
        return self._instruments

    def get_effect_modifiers(self):
        self._effect_modifiers = self._graph.get_effect_modifiers(self._treatment, self._outcome)
        return self._effect_modifiers

    def learn_graph(self, method_name="cdt.causality.graph.LiNGAM", *args, **kwargs):
        """
        Learn causal graph from the data. This function takes the method name as input and initializes the
        causal graph object using the learnt graph.

        :param self: instance of the CausalModel class (or its subclass)
        :param method_name: Exact method name of the object to be imported from the concerned library.
        :returns: an instance of the CausalGraph class initialized with the learned graph.
        """
        # Import causal discovery class
        str_arr = method_name.split(".", maxsplit=1)
        library_name = str_arr[0]
        causal_discovery_class = graph_learners.get_discovery_class_object(library_name)

        model = causal_discovery_class(self._data, method_name, *args, **kwargs)
        graph = model.learn_graph()

        # Initialize causal graph object
        self.init_graph(graph=graph)

        return self._graph

    def identify_effect(
        self, estimand_type=None, method_name="default", proceed_when_unidentifiable=None, optimize_backdoor=False
    ):
        """Identify the causal effect to be estimated, using properties of the causal graph.

        :param method_name: Method name for identification algorithm. ("id-algorithm" or "default")
        :param proceed_when_unidentifiable: Binary flag indicating whether identification should proceed in the presence of (potential) unobserved confounders.
        :returns: a probability expression (estimand) for the causal effect if identified, else NULL

        """
        if proceed_when_unidentifiable is None:
            proceed_when_unidentifiable = self._proceed_when_unidentifiable
        if estimand_type is None:
            estimand_type = self._estimand_type

        estimand_type = EstimandType(estimand_type)

        if method_name == "id-algorithm":
            identifier = IDIdentifier()
        else:
            identifier = AutoIdentifier(
                estimand_type=estimand_type,
                backdoor_adjustment=BackdoorAdjustment(method_name),
                proceed_when_unidentifiable=proceed_when_unidentifiable,
                optimize_backdoor=optimize_backdoor,
            )

        identified_estimand = identifier.identify_effect(
            graph=self._graph, treatment_name=self._treatment, outcome_name=self._outcome
        )

        self.identifier = identifier

        return identified_estimand

    def estimate_effect(
        self,
        identified_estimand,
        method_name=None,
        control_value=0,
        treatment_value=1,
        test_significance=None,
        evaluate_effect_strength=False,
        confidence_intervals=False,
        target_units="ate",
        effect_modifiers=None,
        fit_estimator=True,
        method_params=None,
    ):
        """Estimate the identified causal effect.

        Currently requires an explicit method name to be specified. Method names follow the convention of identification method followed by the specific estimation method: "[backdoor/iv].estimation_method_name". Following methods are supported.
            * Propensity Score Matching: "backdoor.propensity_score_matching"
            * Propensity Score Stratification: "backdoor.propensity_score_stratification"
            * Propensity Score-based Inverse Weighting: "backdoor.propensity_score_weighting"
            * Linear Regression: "backdoor.linear_regression"
            * Generalized Linear Models (e.g., logistic regression): "backdoor.generalized_linear_model"
            * Instrumental Variables: "iv.instrumental_variable"
            * Regression Discontinuity: "iv.regression_discontinuity"

        In addition, you can directly call any of the EconML estimation methods. The convention is "backdoor.econml.path-to-estimator-class". For example, for the double machine learning estimator ("DML" class) that is located inside "dml" module of EconML, you can use the method name, "backdoor.econml.dml.DML". CausalML estimators can also be called. See `this demo notebook <https://py-why.github.io/dowhy/example_notebooks/dowhy-conditional-treatment-effects.html>`_.


        :param identified_estimand: a probability expression
            that represents the effect to be estimated. Output of
            CausalModel.identify_effect method
        :param method_name: name of the estimation method to be used.
        :param control_value: Value of the treatment in the control group, for effect estimation.  If treatment is multi-variate, this can be a list.
        :param treatment_value: Value of the treatment in the treated group, for effect estimation. If treatment is multi-variate, this can be a list.
        :param test_significance: Binary flag on whether to additionally do a statistical signficance test for the estimate.
        :param evaluate_effect_strength: (Experimental) Binary flag on whether to estimate the relative strength of the treatment's effect. This measure can be used to compare different treatments for the same outcome (by running this method with different treatments sequentially).
        :param confidence_intervals: (Experimental) Binary flag indicating whether confidence intervals should be computed.
        :param target_units: (Experimental) The units for which the treatment effect should be estimated. This can be of three types. (1) a string for common specifications of target units (namely, "ate", "att" and "atc"), (2) a lambda function that can be used as an index for the data (pandas DataFrame), or (3) a new DataFrame that contains values of the effect_modifiers and effect will be estimated only for this new data.
        :param effect_modifiers: Names of effect modifier variables can be (optionally) specified here too, since they do not affect identification. If None, the effect_modifiers from the CausalModel are used.
        :param fit_estimator: Boolean flag on whether to fit the estimator.
            Setting it to False is useful to estimate the effect on new data using a previously fitted estimator.
        :param method_params: Dictionary containing any method-specific parameters. These are passed directly to the estimating method. See the docs for each estimation method for allowed method-specific params.
        :returns: An instance of the CausalEstimate class, containing the causal effect estimate
            and other method-dependent information

        """

        if effect_modifiers is None or len(effect_modifiers) == 0:
            effect_modifiers = self._graph.get_effect_modifiers(self._treatment, self._outcome)

        if method_name is None:
            # TODO add propensity score as default backdoor method, iv as default iv method, add an informational message to show which method has been selected.
            pass
        else:
            # TODO add dowhy as a prefix to all dowhy estimators
            num_components = len(method_name.split("."))
            str_arr = method_name.split(".", maxsplit=1)
            identifier_name = str_arr[0]
            estimator_name = str_arr[1]
            # This is done as all dowhy estimators have two parts and external ones have two or more parts
            if num_components > 2:
                estimator_package = estimator_name.split(".")[0]
                if estimator_package == "dowhy":  # For updated dowhy methods
                    estimator_method = estimator_name.split(".", maxsplit=1)[
                        1
                    ]  # discard dowhy from the full package name
                    causal_estimator_class = causal_estimators.get_class_object(estimator_method + "_estimator")
                else:
                    third_party_estimator_package = estimator_package
                    causal_estimator_class = causal_estimators.get_class_object(
                        third_party_estimator_package, estimator_name
                    )
                    if method_params is None:
                        method_params = {}
                    # Define the third-party estimation method to be used
                    method_params[third_party_estimator_package + "_estimator"] = estimator_name
            else:  # For older dowhy methods
                self.logger.info(estimator_name)
                # Process the dowhy estimators
                causal_estimator_class = causal_estimators.get_class_object(estimator_name + "_estimator")

            if method_params is not None and (num_components <= 2 or estimator_package == "dowhy"):
                extra_args = method_params.get("init_params", {})
            else:
                extra_args = {}
            if method_params is None:
                method_params = {}

            identified_estimand.set_identifier_method(identifier_name)

            if not fit_estimator and method_name in self._estimator_cache:
                causal_estimator = self._estimator_cache[method_name]
            else:
                causal_estimator = causal_estimator_class(
                    identified_estimand,
                    **method_params,
                    **extra_args,
                )

                self._estimator_cache[method_name] = causal_estimator

        return estimate_effect(
            self._data,
            self._treatment,
            self._outcome,
            identifier_name,
            causal_estimator,
            control_value,
            treatment_value,
            target_units,
            effect_modifiers,
            fit_estimator,
            method_params,
        )

    def do(self, x, identified_estimand, method_name=None, fit_estimator=True, method_params=None):
        """Do operator for estimating values of the outcome after intervening on treatment.

        :param x: interventional value of the treatment variable
        :param identified_estimand: a probability expression
            that represents the effect to be estimated. Output of
            CausalModel.identify_effect method
        :param method_name: any of the estimation method to be used. See docs
            for estimate_effect method for a list of supported estimation methods.
        :param fit_estimator: Boolean flag on whether to fit the estimator.
            Setting it to False is useful to compute the do-operation on new
            data using a previously fitted estimator.
        :param method_params: Dictionary containing any method-specific parameters. These are passed directly to the estimating method.

        :returns: an instance of the CausalEstimate class, containing the causal effect estimate
            and other method-dependent information

        """
        if method_name is None:
            pass
        else:
            str_arr = method_name.split(".", maxsplit=1)
            identifier_name = str_arr[0]
            estimator_name = str_arr[1]
            identified_estimand.set_identifier_method(identifier_name)
            causal_estimator_class = causal_estimators.get_class_object(estimator_name + "_estimator")

        # Check if estimator's target estimand is identified
        if identified_estimand.estimands[identifier_name] is None:
            self.logger.warning("No valid identified estimand for using instrumental variables method")
            estimate = CausalEstimate(None, None, None, None, None)
        else:
            if fit_estimator:
                # Note that while the name of the variable is the same,
                # "self.causal_estimator", this estimator takes in less
                # parameters than the same from the
                # estimate_effect code. It is not advisable to use the
                # estimator from this function to call estimate_effect
                # with fit_estimator=False.
                self.causal_estimator = causal_estimator_class(
                    identified_estimand,
                    **method_params,
                )
                self.causal_estimator.fit(
                    self._data,
                    self._treatment,
                    self._outcome,
                )
            else:
                # Estimator had been computed in a previous call
                assert self.causal_estimator is not None
            try:
                estimate = self.causal_estimator.do(x)
            except NotImplementedError:
                self.logger.error("Do Operation not implemented or not supported for this estimator.")
                raise NotImplementedError
        return estimate

    def refute_estimate(self, estimand, estimate, method_name=None, show_progress_bar=False, **kwargs):
        """Refute an estimated causal effect.

        If method_name is provided, uses the provided method. In the future, we may support automatic selection of suitable refutation tests. Following refutation methods are supported.
            * Adding a randomly-generated confounder: "random_common_cause"
            * Adding a confounder that is associated with both treatment and outcome: "add_unobserved_common_cause"
            * Replacing the treatment with a placebo (random) variable): "placebo_treatment_refuter"
            * Removing a random subset of the data: "data_subset_refuter"

        :param estimand: target estimand, an instance of the IdentifiedEstimand class (typically, the output of identify_effect)
        :param estimate: estimate to be refuted, an instance of the CausalEstimate class (typically, the output of estimate_effect)
        :param method_name: name of the refutation method
        :param show_progress_bar: Boolean flag on whether to show a progress bar
        :param kwargs:  (optional) additional arguments that are passed directly to the refutation method. Can specify a random seed here to ensure reproducible results ('random_seed' parameter). For method-specific parameters, consult the documentation for the specific method. All refutation methods are in the causal_refuters subpackage.

        :returns: an instance of the RefuteResult class

        """
        if estimate is None or estimate.value is None:
            self.logger.error("Aborting refutation! No estimate is provided.")
            raise ValueError("Aborting refutation! No valid estimate is provided.")
        if method_name is None:
            pass
        else:
            refuter_class = causal_refuters.get_class_object(method_name)

        refuter = refuter_class(self._data, identified_estimand=estimand, estimate=estimate, **kwargs)
        res = refuter.refute_estimate(show_progress_bar)
        return res

    def view_model(self, layout="dot", size=(8, 6), file_name="causal_model"):
        """View the causal DAG.

        :param layout: string specifying the layout of the graph.
        :param size: tuple (x, y) specifying the width and height of the figure in inches.
        :param file_name: string specifying the file name for the saved causal graph png.

        :returns: a visualization of the graph

        """
        self._graph.view_graph(layout, size, file_name)

    def interpret(self, method_name=None, **kwargs):
        """Interpret the causal model.

        :param method_name: method used for interpreting the model. If None,
                            then default interpreter is chosen that describes the model summary and shows the associated causal graph.
        :param kwargs:: Optional parameters that are directly passed to the interpreter method.

        :returns: None

        """
        if method_name is None:
            self.summary(print_to_stdout=True)
            self.view_model()
            return

        method_name_arr = parse_state(method_name)
        import dowhy.interpreters as interpreters

        for method in method_name_arr:
            interpreter = interpreters.get_class_object(method)
            interpreter(self, **kwargs).interpret()

    def summary(self, print_to_stdout=False):
        """Print a text summary of the model.

        :returns: a string containining the summary

        """
        summary_text = "Model to find the causal effect of treatment {0} on outcome {1}".format(
            self._treatment, self._outcome
        )
        self.logger.info(summary_text)
        if print_to_stdout:
            print(summary_text)
        return summary_text

    def refute_graph(self, k=1, independence_test=None, independence_constraints=None):
        """
        Check if the dependencies in input graph matches with the dataset -
        ( X ⫫ Y ) | Z
        where X and Y are considered as singleton sets currently
        Z can have multiple variables
        :param k: number of covariates in set Z
        :param independence_test: dictionary containing methods to test conditional independece in data
        :param independence_constraints: list of implications to be test input by the user in the format
            [(x,y,(z1,z2)),
            (x,y, (z3,))
            ]
        : returns: an instance of GraphRefuter class
        """
        if independence_test is not None:
            test_for_continuous = independence_test["test_for_continuous"]
            test_for_discrete = independence_test["test_for_discrete"]
            refuter = GraphRefuter(
                data=self._data, method_name_continuous=test_for_continuous, method_name_discrete=test_for_discrete
            )

        else:
            refuter = GraphRefuter(data=self._data)

        if independence_constraints is None:
            all_nodes = list(self._graph.get_all_nodes(include_unobserved=False))
            num_nodes = len(all_nodes)
            array_indices = list(range(0, num_nodes))
            all_possible_combinations = list(
                combinations(array_indices, 2)
            )  # Generating sets of indices of size 2 for different x and y
            conditional_independences = []
            self.logger.info("The followed conditional independences are true for the input graph")
            for combination in all_possible_combinations:  # Iterate over the unique 2-sized sets [x,y]
                i = combination[0]
                j = combination[1]
                a = all_nodes[i]
                b = all_nodes[j]
                if i < j:
                    temp_arr = all_nodes[:i] + all_nodes[i + 1 : j] + all_nodes[j + 1 :]
                else:
                    temp_arr = all_nodes[:j] + all_nodes[j + 1 : i] + all_nodes[i + 1 :]
                k_sized_lists = list(combinations(temp_arr, k))
                for k_list in k_sized_lists:
                    if self._graph.check_dseparation([str(a)], [str(b)], k_list) == True:
                        self.logger.info(" %s and %s are CI given %s ", a, b, k_list)
                        conditional_independences.append([a, b, k_list])

            independence_constraints = conditional_independences

        res = refuter.refute_model(independence_constraints=independence_constraints)

        self.logger.info(refuter._refutation_passed)

        return res



class OrderedSet:
    """
    Python class for ordered set.
    Code taken from https://github.com/buyalsky/ordered-hash-set/tree/5198b23e01faeac3f5398ab2c08cb013d14b3702.
    """

    def __init__(self, elements=None):
        self._set = {}
        self._start = None
        self._end = None
        if elements is not None:
            for element in elements:
                self.add(element)

    def add(self, element):
        """
        Function to add an element to do set if it does not exit.

        :param element: element to be added.
        """
        if self._start is None:
            self._start = element

        if element not in self._set.keys():
            self._set[element] = None
            if len(self._set) > 1:
                self._set[self._end] = element
            self._end = element

    def get_all(self):
        """
        Function to return list of all elements in the set.

        :returns: List of all items in the set.
        """
        return list(self)

    def is_empty(self):
        """
        Function to determine if the set is empty or not.

        :returns: ``True`` if the set is empty, ``False`` otherwise.
        """
        return self.__len__() == 0

    def union(self, other_set):
        """
        Function to compute the union of self._set and other_set.

        :param other_set: The set to obtain union with. Can be a list, set or OrderedSet.
        :returns: New OrderedSet representing the set with elements from the OrderedSet object and other_set.
        """
        new_set = OrderedSet()
        for element in self._set:
            new_set.add(element)
        for element in other_set:
            new_set.add(element)
        return new_set

    def intersection(self, other_set):
        """
        Function to compute the intersection of self._set and other_set.

        :param other_set: The set to obtain intersection with. Can be a list, set or OrderedSet.
        :returns: New OrderedSet representing the set with elements common to the OrderedSet object and other_set.
        """
        new_set = OrderedSet()
        for element in self._set:
            if element in other_set:
                new_set.add(element)
        return new_set

    def difference(self, other_set):
        """
        Function to remove elements in self._set which are also present in other_set.

        :param other_set: The set to obtain difference with. Can be a list, set or OrderedSet.
        :returns: New OrderedSet representing the difference of elements in the self._set and other_set.
        """
        new_set = OrderedSet()
        for element in self._set:
            if element not in other_set:
                new_set.add(element)
        return new_set

    def __getitem__(self, index):
        if index >= self.__len__():
            raise IndexError("Index is out of range")
        return list(self)[index]

    def __iter__(self):
        self._iter = self._start
        return self

    def __next__(self):
        element = self._iter
        if not element:
            raise StopIteration
        self._iter = self._set[element]
        return element

    def __len__(self):
        return len(self._set)

    def __str__(self):
        elements = [str(i) for i in self]
        string = "OrderedSet(" + ",".join(elements) + ")"
        return string

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return False

        return self._set == other._set



import networkx as nx
import numpy as np
import pandas as pd
import pytest
from flaky import flaky
from pytest import approx

from dowhy.gcm import (
    AdditiveNoiseModel,
    ClassifierFCM,
    EmpiricalDistribution,
    InvertibleStructuralCausalModel,
    ProbabilisticCausalModel,
    auto,
    average_causal_effect,
    counterfactual_samples,
    fit,
    interventional_samples,
)
from dowhy.gcm.ml import create_linear_regressor, create_logistic_regression_classifier


def __create_and_fit_simple_probabilistic_causal_model():
    X0 = np.random.uniform(-1, 1, 10000)
    X1 = 2 * X0 + np.random.normal(0, 0.1, 10000)
    X2 = 0.5 * X0 + np.random.normal(0, 0.1, 10000)
    X3 = 0.5 * X2 + np.random.normal(0, 0.1, 10000)

    original_observations = pd.DataFrame({"X0": X0, "X1": X1, "X2": X2, "X3": X3})

    causal_model = InvertibleStructuralCausalModel(nx.DiGraph([("X0", "X1"), ("X0", "X2"), ("X2", "X3")]))
    causal_model.set_causal_mechanism("X0", EmpiricalDistribution())
    causal_model.set_causal_mechanism("X1", AdditiveNoiseModel(prediction_model=create_linear_regressor()))
    causal_model.set_causal_mechanism("X2", AdditiveNoiseModel(prediction_model=create_linear_regressor()))
    causal_model.set_causal_mechanism("X3", AdditiveNoiseModel(prediction_model=create_linear_regressor()))

    fit(causal_model, original_observations)

    return causal_model, original_observations


@flaky(max_runs=3)
def test_interventional_samples_atomic():
    causal_model, _ = __create_and_fit_simple_probabilistic_causal_model()

    observed_data = pd.DataFrame({"X0": [0], "X1": [1], "X2": [2], "X3": [3]})

    sample = interventional_samples(causal_model, dict(X2=lambda x: np.array(10)), observed_data).to_numpy()
    sample = sample.squeeze()
    assert sample[0] == 0
    assert sample[1] == 1
    assert sample[2] == 10
    assert sample[3] == approx(5, abs=0.2)


@flaky(max_runs=3)
def test_interventional_samples_conditional():
    causal_model, _ = __create_and_fit_simple_probabilistic_causal_model()

    observed_data = pd.DataFrame({"X0": [0], "X1": [1], "X2": [2], "X3": [3]})

    sample = interventional_samples(causal_model, dict(X2=lambda x: x + 10), observed_data).to_numpy()
    sample = sample.squeeze()
    assert sample[0] == 0
    assert sample[1] == 1
    assert sample[2] == approx(10, abs=0.5)
    assert sample[3] == approx(5, abs=0.25)


@flaky(max_runs=3)
def test_interventional_samples_atomic_draw():
    causal_model, _ = __create_and_fit_simple_probabilistic_causal_model()

    samples = interventional_samples(causal_model, dict(X2=lambda x: np.array(10)), num_samples_to_draw=10)
    assert samples["X2"].to_numpy() == approx(10, abs=0)


@flaky(max_runs=3)
def test_interventional_samples_conditional_draw():
    causal_model, training_data = __create_and_fit_simple_probabilistic_causal_model()

    samples = interventional_samples(causal_model, dict(X2=lambda x: x + 10), num_samples_to_draw=10)
    assert samples["X2"].to_numpy() == approx(np.mean(training_data["X2"].to_numpy()) + 10, abs=1)


@flaky(max_runs=3)
def test_interventional_samples_with_categorical_variables_draw():
    causal_model = ProbabilisticCausalModel(nx.DiGraph([("X0", "X2"), ("X1", "X2"), ("X2", "X3")]))
    causal_model.set_causal_mechanism("X0", EmpiricalDistribution())
    causal_model.set_causal_mechanism("X1", EmpiricalDistribution())
    causal_model.set_causal_mechanism("X2", ClassifierFCM(classifier_model=create_logistic_regression_classifier()))
    causal_model.set_causal_mechanism("X3", ClassifierFCM(classifier_model=create_logistic_regression_classifier()))

    X0 = np.random.normal(0, 1, 5000)
    X1 = np.random.choice(2, 5000).astype(str)

    X2 = []
    for (x0, x1) in zip(X0, X1):
        if x1 == "0":
            X2.append(x0 + 2 > 0)
        else:
            X2.append(x0 - 2 > 0)

    X2 = np.array(X2).astype(str)

    X3 = []
    for x2 in X2:
        if x2 == "True":
            X3.append("False")
        else:
            X3.append("True")

    X3 = np.array(X3).astype(str)

    training_data = pd.DataFrame({"X0": X0, "X1": X1, "X2": X2, "X3": X3})

    fit(causal_model, training_data)

    intervention_results = interventional_samples(
        causal_model, dict(X0=lambda x: 0, X1=lambda x: "0"), num_samples_to_draw=100
    )

    assert intervention_results["X0"].to_numpy() == approx(np.array([0] * 100))
    assert np.all(intervention_results["X1"].to_numpy() == np.array(["0"] * 100))
    assert np.sum(intervention_results["X2"].to_numpy() == "True") > 95
    assert np.sum(intervention_results["X3"].to_numpy() == "False") > 95


@flaky(max_runs=3)
def test_interventional_samples_atomic_multiple_interventions():
    causal_model, _ = __create_and_fit_simple_probabilistic_causal_model()

    observed_data = pd.DataFrame({"X0": [0], "X1": [1], "X2": [2], "X3": [3]})

    sample = interventional_samples(causal_model, dict(X0=lambda x: 10, X2=lambda x: x + 5), observed_data).to_numpy()
    sample = sample.squeeze()
    assert sample[0] == 10
    assert sample[1] == approx(20, abs=0.3)
    assert sample[2] == approx(10, abs=0.3)
    assert sample[3] == approx(5, abs=0.3)


def test_interventional_samples_raise_error_all_parameter_none():
    causal_model, _ = __create_and_fit_simple_probabilistic_causal_model()

    with pytest.raises(ValueError):
        interventional_samples(causal_model, dict(X0=lambda x: 10))


def test_interventional_samples_raise_error_both_parameter_given():
    causal_model, _ = __create_and_fit_simple_probabilistic_causal_model()

    observed_data = pd.DataFrame({"X0": [0], "X1": [1], "X2": [2], "X3": [3]})

    with pytest.raises(ValueError):
        interventional_samples(
            causal_model, dict(X0=lambda x: 10), observed_data=observed_data, num_samples_to_draw=100
        )


def test_counterfactual_samples_with_observed_samples():
    causal_model, _ = __create_and_fit_simple_probabilistic_causal_model()

    observed_samples = pd.DataFrame({"X0": [1], "X1": [3], "X2": [3], "X3": [4]})

    sample = counterfactual_samples(causal_model, dict(X2=lambda x: 2), observed_data=observed_samples)
    assert sample["X0"].to_numpy().squeeze() == 1
    assert sample["X1"].to_numpy().squeeze() == 3
    assert sample["X2"].to_numpy().squeeze() == 2
    assert sample["X3"].to_numpy().squeeze() == approx(3.5, abs=0.05)


def test_counterfactual_samples_with_noise_samples():
    causal_model, _ = __create_and_fit_simple_probabilistic_causal_model()

    noise_samples = pd.DataFrame({"X0": [1], "X1": [2], "X2": [3], "X3": [4]})

    sample = counterfactual_samples(causal_model, dict(X2=lambda x: 2), noise_data=noise_samples)
    assert sample["X0"].to_numpy().squeeze() == 1
    assert sample["X1"].to_numpy().squeeze() == approx(4, abs=0.05)
    assert sample["X2"].to_numpy().squeeze() == 2
    assert sample["X3"].to_numpy().squeeze() == approx(5, abs=0.05)


def test_counterfactual_samples_raises_error_all_parameter_none():
    causal_model, _ = __create_and_fit_simple_probabilistic_causal_model()

    with pytest.raises(ValueError):
        counterfactual_samples(causal_model, dict(X0=lambda x: 10))


def test_counterfactual_samples_raises_error_both_parameter_given():
    causal_model, _ = __create_and_fit_simple_probabilistic_causal_model()

    with pytest.raises(ValueError):
        counterfactual_samples(
            causal_model,
            dict(X0=lambda x: 10),
            observed_data=pd.DataFrame({"X0": [1], "X1": [3], "X2": [3], "X3": [4]}),
            noise_data=pd.DataFrame({"X0": [1], "X1": [4], "X2": [3], "X3": [4]}),
        )


@flaky(max_runs=3)
def test_given_continuous_target_when_estimate_average_causal_effect_then_return_expected_result():
    T = np.random.choice(2, 1000, replace=True)
    X0 = np.random.normal(0, 0.2, 1000) + T
    X1 = np.random.normal(0, 0.2, 1000) + 0.5 * T
    Y = X0 + X1 + np.random.normal(0, 0.1, 1000)

    data = pd.DataFrame(dict(T=T, X0=X0, X1=X1, Y=Y))

    causal_model = ProbabilisticCausalModel(nx.DiGraph([("T", "X0"), ("T", "X1"), ("X0", "Y"), ("X1", "Y")]))
    auto.assign_causal_mechanisms(causal_model, data, auto.AssignmentQuality.GOOD)
    fit(causal_model, data)

    assert average_causal_effect(
        causal_model,
        "Y",
        interventions_alternative={"T": lambda x: 1},
        interventions_reference={"T": lambda x: 0},
        num_samples_to_draw=1000,
    ) == approx(1.5, abs=0.1)


@flaky(max_runs=3)
def test_given_binary_target_when_estimate_average_causal_effect_then_return_expected_result():
    T = np.random.choice(2, 1000, replace=True)
    X0 = np.random.normal(0, 0.1, 1000) + T
    X1 = np.random.normal(0, 0.1, 1000) + 0.5 * T
    Y = ((X0 + X1 + np.random.normal(0, 0.1, 1000)) >= 1.5).astype(str)

    data = pd.DataFrame(dict(T=T, X0=X0, X1=X1, Y=Y))

    causal_model = ProbabilisticCausalModel(nx.DiGraph([("T", "X0"), ("T", "X1"), ("X0", "Y"), ("X1", "Y")]))
    auto.assign_causal_mechanisms(causal_model, data, auto.AssignmentQuality.GOOD)
    fit(causal_model, data)

    assert average_causal_effect(
        causal_model,
        "Y",
        interventions_alternative={"T": lambda x: 1},
        interventions_reference={"T": lambda x: 0},
        num_samples_to_draw=1000,
    ) == approx(0.5, abs=0.1)



import numpy as np
from flaky import flaky
from pytest import approx
from scipy.stats import entropy

from dowhy.gcm.uncertainty import (
    estimate_entropy_discrete,
    estimate_entropy_kmeans,
    estimate_entropy_of_probabilities,
    estimate_entropy_using_discretization,
    estimate_gaussian_entropy,
    estimate_variance,
)


def test_estimate_entropy_using_discretization():
    X = np.random.normal(0, 1, 10000)
    Y = np.random.normal(0, 3, 10000)

    assert estimate_entropy_using_discretization(X) == approx(1.25, abs=0.2)
    assert estimate_entropy_using_discretization(Y) == approx(2.45, abs=0.2)


def test_estimate_entropy_kmeans():
    X = np.random.normal(0, 1, 10000)
    Y = np.random.normal(0, 3, 10000)

    assert estimate_entropy_kmeans(X) == approx(1.4, abs=0.2)
    assert estimate_entropy_kmeans(Y) == approx(2.5, abs=0.2)


def test_estimate_gaussian_entropy():
    X = np.random.normal(0, 1, 10000)
    Y = np.random.normal(0, 3, 10000)

    assert estimate_gaussian_entropy(X) == approx(1.4, abs=0.2)
    assert estimate_gaussian_entropy(Y) == approx(2.5, abs=0.2)


@flaky(max_runs=5)
def test_estimate_variance():
    X = np.random.normal(0, 1, 10000)
    Y = np.random.normal(0, 3, 10000)

    assert estimate_variance(X) == approx(1, abs=0.2)
    assert estimate_variance(Y) == approx(9, abs=0.2)

    X = np.random.normal(0, 1, (1000000, 3))
    Y = np.random.normal(0, 3, (1000000, 3))

    assert estimate_variance(X) == approx(1, abs=1)
    assert estimate_variance(Y) == approx(729, abs=15)


def test_estimate_entropy_of_probabilities():
    assert estimate_entropy_of_probabilities(np.array([[0.25, 0.25, 0.25, 0.25]])) == approx(1.38, abs=0.05)
    assert estimate_entropy_of_probabilities(np.array([[1, 0, 0, 0]])) == approx(0, abs=0.05)
    assert estimate_entropy_of_probabilities(np.array([[0.5, 0, 0, 0.5]])) == approx(0.69, abs=0.05)
    assert estimate_entropy_of_probabilities(np.array([[0, 0.5, 0.5, 0], [0.5, 0, 0, 0.5]])) == approx(0.69, abs=0.05)
    assert estimate_entropy_of_probabilities(np.array([[1, 0, 0, 0], [1, 0, 0, 0]])) == approx(0, abs=0.05)


def test_estimate_entropy_discrete():
    assert estimate_entropy_discrete(np.random.choice(2, (1000, 1), replace=True)) == approx(entropy([0.5, 0.5]), 0.05)
    assert estimate_entropy_discrete(np.random.choice(2, (1000, 2), replace=True)) == approx(
        entropy([0.25, 0.25, 0.25, 0.25]), 0.05
    )



import numpy as np
import pytest

from dowhy.gcm.density_estimators import GaussianMixtureDensityEstimator, KernelDensityEstimator1D


def test_when_fit_and_evaluate_gaussian_mixture_density_estimator_then_behaves_as_expected():
    test_data = np.array([[0, 1], [0, 2], [1, 0], [2, 3]])

    density_estimator_model = GaussianMixtureDensityEstimator()
    with pytest.raises(RuntimeError):
        density_estimator_model.density(test_data)

    density_estimator_model.fit(test_data)
    results = density_estimator_model.density(test_data)

    assert len(results) == 4


def test_when_fit_and_evaluate_kernel_based_density_estimator_1d_then_behaves_as_expected():
    test_data = np.array([[0, 1], [0, 2], [1, 0], [2, 3]])

    density_estimator_model = KernelDensityEstimator1D()
    with pytest.raises(RuntimeError):
        density_estimator_model.density(test_data)

    with pytest.raises(RuntimeError):
        density_estimator_model.fit(test_data)

    test_data = np.array([[0], [2], [1], [3]])
    density_estimator_model.fit(test_data)
    results = density_estimator_model.density(test_data)

    assert len(results) == 4



from typing import Any, Callable, Dict, Optional, Union

import numpy as np
import pandas as pd
from numpy.matlib import repmat
from tqdm import tqdm

from dowhy.gcm import config
from dowhy.gcm._noise import compute_noise_from_data, get_noise_dependent_function, noise_samples_of_ancestors
from dowhy.gcm.anomaly_scorer import AnomalyScorer
from dowhy.gcm.anomaly_scorers import MedianCDFQuantileScorer, RescaledMedianCDFQuantileScorer
from dowhy.gcm.cms import InvertibleStructuralCausalModel, ProbabilisticCausalModel
from dowhy.gcm.graph import ConditionalStochasticModel, get_ordered_predecessors, is_root_node, validate_causal_dag
from dowhy.gcm.shapley import ShapleyConfig, estimate_shapley_values
from dowhy.gcm.stats import permute_features
from dowhy.gcm.util.general import shape_into_2d


def conditional_anomaly_scores(
    parent_samples: np.ndarray,
    target_samples: np.ndarray,
    causal_mechanism: ConditionalStochasticModel,
    anomaly_scorer_factory: Callable[[], AnomalyScorer] = MedianCDFQuantileScorer,
    num_samples_conditional: int = 10000,
) -> np.ndarray:
    """Estimates the conditional anomaly scores based on the expected outcomes of the causal model.

    :param parent_samples: Samples from all parents of the target node.
    :param target_samples: Samples from the target node.
    :param causal_mechanism: Causal mechanism of the target node.
    :param anomaly_scorer_factory: A callable that returns an anomaly scorer.
    :param num_samples_conditional: Number of samples drawn from the conditional distribution based on the given
                                    parent samples. The more samples, the more accurate the results.
    :return: The conditional anomaly score for each sample in target_samples.
    """
    parent_samples, target_samples = shape_into_2d(parent_samples, target_samples)
    if parent_samples.shape[0] != target_samples.shape[0]:
        raise ValueError("There should be as many parent samples as target samples!")

    result = np.zeros(parent_samples.shape[0])
    for i in range(parent_samples.shape[0]):
        samples_from_conditional = causal_mechanism.draw_samples(repmat(parent_samples[i], num_samples_conditional, 1))
        anomaly_scorer = anomaly_scorer_factory()
        anomaly_scorer.fit(samples_from_conditional)
        result[i] = anomaly_scorer.score(target_samples[i])[0]

    return result


def anomaly_scores(
    causal_model: ProbabilisticCausalModel,
    anomaly_data: pd.DataFrame,
    num_samples_conditional: int = 10000,
    num_samples_unconditional: int = 10000,
    anomaly_scorer_factory: Callable[[], AnomalyScorer] = RescaledMedianCDFQuantileScorer,
) -> Dict[Any, np.ndarray]:
    if isinstance(anomaly_data, pd.Series):
        anomaly_data = pd.DataFrame([anomaly_data])

    validate_causal_dag(causal_model.graph)

    results = {}
    for node in tqdm(
        causal_model.graph.nodes,
        desc="Estimating conditional anomaly scores",
        position=0,
        leave=True,
        disable=not config.show_progress_bars,
    ):

        if is_root_node(causal_model.graph, node):
            anomaly_scorer = anomaly_scorer_factory()
            anomaly_scorer.fit(causal_model.causal_mechanism(node).draw_samples(num_samples_unconditional))
            results[node] = anomaly_scorer.score(anomaly_data[node].to_numpy())
        else:
            tmp_anomaly_parent_samples = anomaly_data[get_ordered_predecessors(causal_model.graph, node)].to_numpy()
            tmp_anomaly_target_samples = anomaly_data[node].to_numpy()
            results[node] = conditional_anomaly_scores(
                tmp_anomaly_parent_samples,
                tmp_anomaly_target_samples,
                causal_model.causal_mechanism(node),
                anomaly_scorer_factory,
                num_samples_conditional,
            )

    return results


def attribute_anomalies(
    causal_model: InvertibleStructuralCausalModel,
    target_node: Any,
    anomaly_samples: pd.DataFrame,
    anomaly_scorer: Optional[AnomalyScorer] = None,
    attribute_mean_deviation: bool = False,
    num_distribution_samples: int = 5000,
    shapley_config: Optional[ShapleyConfig] = None,
) -> Dict[Any, np.ndarray]:
    """Estimates the contributions of upstream nodes to the anomaly score of the target_node for each sample in
    anomaly_samples. By default, the anomaly score is based on the information theoretic (IT) score
    -log(P(g(X) >= g(x))), where g is the anomaly_scorer, X samples from the marginal
    distribution of the target_node and x an observation of the target_node in anomaly_samples. If
    attribute_mean_deviation is set to True, the contribution to g(x) - E[g(X)] is estimated instead, i.e. the feature
    relevance for the given scoring function. The underlying algorithm utilizes the reconstructed noise of upstream
    nodes (including the target_node itself) for the given anomaly_samples. By this, it is possible to estimate how
    much of the anomaly score can be explained by upstream anomalies with respect to anomalous noise values.

    Note: This function requires that the noise can be recovered from samples, i.e. the causal models of non-root nodes
    need to be an InvertibleNoiseModel (e.g. AdditiveNoiseModel).

    Related paper:
    Janzing, D., Budhathoki, K., Minorics, L., & Bloebaum, P. (2019).
    Causal structure based root cause analysis of outliers
    https://arxiv.org/abs/1912.02724

    :param causal_model: The fitted InvertibleStructuralCausalModel.
    :param target_node: Target node for which the contributions are estimated.
    :param anomaly_samples: Anomalous observations for which the contributions are estimated.
    :param anomaly_scorer: Anomaly scorer g. If None is given, a MedianCDFQuantileScorer is used.
    :param attribute_mean_deviation: If set to False, the contribution is estimated based on the IT score and if it is
                                     set to True, the contribution is based on the feature relevance with respect to the given scoring function.
    :param num_distribution_samples: Number of samples from X, the marginal distribution of the target. These are used
                                     for evaluating the tail probability in case of the IT score
                                     (attribute_mean_deviation is False) or as samples for randomization in case of
                                     feature relevance (attribute_mean_deviation is True).
    :param shapley_config: :class:`~dowhy.gcm.shapley.ShapleyConfig` for the Shapley estimator.
    :return: A dictionary that assigns a numpy array to each upstream node including the target_node itself. The
             i-th entry of an array indicates the contribution of the corresponding node to the anomaly score of the target
             for the i-th observation in anomaly_samples.
    """
    validate_causal_dag(causal_model.graph)

    if anomaly_scorer is None:
        anomaly_scorer = MedianCDFQuantileScorer()

    noise_of_anomaly_samples = compute_noise_from_data(causal_model, anomaly_samples)
    node_samples, noise_samples = noise_samples_of_ancestors(causal_model, target_node, num_distribution_samples)
    noise_dependent_function, nodes_order = get_noise_dependent_function(causal_model, target_node)
    anomaly_scorer.fit(node_samples[target_node].to_numpy())

    attributions = attribute_anomaly_scores(
        noise_of_anomaly_samples[nodes_order].to_numpy(),
        noise_samples[nodes_order].to_numpy(),
        lambda x: anomaly_scorer.score(noise_dependent_function(x)),
        attribute_mean_deviation,
        shapley_config,
    )

    return {node: attributions[:, i] for i, node in enumerate(nodes_order)}


def attribute_anomaly_scores(
    anomaly_samples: np.ndarray,
    distribution_samples: np.ndarray,
    anomaly_scoring_func: Callable[[np.ndarray], np.ndarray],
    attribute_mean_deviation: bool,
    shapley_config: Optional[ShapleyConfig] = None,
) -> np.ndarray:
    """Estimates the contributions of the features for each sample in anomaly_samples to the anomaly score obtained
    by the anomaly_scoring_func. If attribute_mean_deviation is set to False, the anomaly score is based on the
    information theoretic (IT) score -log(P(g(X) >= g(x))), where g is the anomaly_scoring_func, X samples from the
    marginal distribution of the target_node and x an observation of the target_node in anomaly_samples. If
    attribute_mean_deviation is set to True, the contribution to g(x) - E[g(X)] is estimated instead, i.e. the
    feature relevance for the given scorer.

    Note that the anomaly scoring function needs to handle the dimension and modality of the data. An example for a
    function for multidimensional continues data would be:
        density_estimator = GaussianMixtureDensityEstimator()
        density_estimator.fit(original_observations)
        anomaly_scoring_func = lambda x, y: estimate_inverse_density_score(x, y, density_estimator)

    Related paper:
    Janzing, D., Budhathoki, K., Minorics, L., & Bloebaum, P. (2022).
    Causal structure based root cause analysis of outliers
    https://arxiv.org/abs/1912.02724

    :param anomaly_samples: Samples x for which the contributions are estimated. The dimensionality of these samples
                            doesn't matter as long as the anomaly_scoring_func supports it.
    :param distribution_samples: Samples from the (non-anomalous) distribution X.
    :param anomaly_scoring_func: A function g that takes a sample from X as input and returns an anomaly score.
    :param attribute_mean_deviation: If set to False, the contribution is estimated based on the IT score and if it is
                                     set to True, the contribution is based on the feature relevance with respect to the
                                     given scoring function.
    :param shapley_config: :class:`~dowhy.gcm.shapley.ShapleyConfig` for the Shapley estimator.
    :return: A numpy array with the feature contributions to the anomaly score for each sample in anomaly_samples.
    """
    if attribute_mean_deviation:
        expectation_of_score = np.mean(anomaly_scoring_func(distribution_samples))
    else:
        anomaly_scores = anomaly_scoring_func(anomaly_samples)

    def set_function(subset: np.ndarray) -> Union[np.ndarray, float]:
        feature_samples = permute_features(distribution_samples, np.arange(0, subset.shape[0])[subset == 0], True)

        result = np.zeros(anomaly_samples.shape[0])
        for i in range(anomaly_samples.shape[0]):
            feature_samples[:, subset == 1] = anomaly_samples[i, subset == 1]

            if attribute_mean_deviation:
                # Usual feature relevance using the mean deviation as set function, i.e. g(x) - E[g(X)]
                result[i] = np.mean(anomaly_scoring_func(feature_samples)) - expectation_of_score
            else:
                result[i] = np.log(_relative_frequency(anomaly_scoring_func(feature_samples) >= anomaly_scores[i]))

        return result

    return estimate_shapley_values(set_function, anomaly_samples.shape[1], shapley_config)


def _relative_frequency(conditions: np.ndarray):
    return (np.sum(conditions) + 0.5) / (len(conditions) + 0.5)



import logging

import numpy as np
from pandas import get_dummies
from sklearn.ensemble import GradientBoostingClassifier
from sklearn.linear_model import LogisticRegression
from sklearn.preprocessing import LabelEncoder
from statsmodels.nonparametric.kernel_density import EstimatorSettings, KDEMultivariateConditional

import dowhy.utils.api as api


def propensity_of_treatment_score(data, covariates, treatment, model="logistic", variable_types=None):
    if model == "logistic":
        model = LogisticRegression(solver="lbfgs")
        data, covariates = binarize_discrete(data, covariates, variable_types)
        model = model.fit(data[covariates], data[treatment].values.ravel())
        scores = model.predict_proba(data[covariates])[:, 1]
        return scores
    else:
        raise NotImplementedError


def state_propensity_score(data, covariates, treatments, variable_types=None):
    if len(set(covariates).intersection(treatments)) != 0:
        raise Exception("Can't control for causal states. Remove treatment from covariates.")
    log_propensities = {}
    for i, treatment in enumerate(treatments):
        if variable_types[treatment] in ["b"]:
            log_propensities[treatment] = np.log(
                binary_treatment_model(data.copy(), covariates + treatments[i + 1 :], treatment, variable_types)
            )
        elif variable_types[treatment] in ["o", "u", "d"]:
            log_propensities[treatment] = np.log(
                categorical_treatment_model(data.copy(), covariates + treatments[i + 1 :], treatment, variable_types)
            )
        elif variable_types[treatment] in ["c"]:
            log_propensities[treatment] = np.log(
                continuous_treatment_model(data.copy(), covariates + treatments[i + 1 :], treatment, variable_types)
            )
        else:
            raise Exception(
                "Variable type {} for variable {} is not a recognized format type.".format(
                    variable_types[treatment], treatment
                )
            )
    scores = np.zeros(len(data))
    for treatment in treatments:
        scores += log_propensities[treatment]
    return np.exp(scores)


def binary_treatment_model(data, covariates, treatment, variable_types):
    data, covariates = binarize_discrete(data, covariates, variable_types)
    model = LogisticRegression(solver="lbfgs")
    model = model.fit(data[covariates], data[treatment])
    scores = model.predict_proba(data[covariates])
    scores = scores[range(len(scores)), data[treatment].values.astype(int)]
    return scores


def categorical_treatment_model(data, covariates, treatment, variable_types):
    data, covariates = binarize_discrete(data, covariates, variable_types)
    model = LogisticRegression(multi_class="ovr", solver="lbfgs")
    data[treatment], encoder = discrete_to_integer(data[treatment])
    model = model.fit(data[covariates], data[treatment])
    scores = model.predict_proba(data[covariates])
    scores = scores[range(len(data)), data[treatment].values.astype(int)]
    return scores


def continuous_treatment_model(data, covariates, treatment, variable_types):
    data, covariates = binarize_discrete(data, covariates, variable_types)
    if len(data) > 300 or len([treatment] + covariates) >= 3:
        defaults = EstimatorSettings(n_jobs=4, efficient=True)
    else:
        defaults = EstimatorSettings(n_jobs=-1, efficient=False)

    if "c" not in variable_types.values():
        bw = "cv_ml"
    else:
        bw = "normal_reference"

    indep_type = get_type_string(covariates, variable_types)
    dep_type = get_type_string([treatment], variable_types)

    model = KDEMultivariateConditional(
        endog=data[treatment],
        exog=data[covariates],
        dep_type="".join(dep_type),
        indep_type="".join(indep_type),
        bw=bw,
        defaults=defaults,
    )
    scores = model.pdf(endog_predict=data[treatment], exog_predict=data[covariates])
    return scores


def get_type_string(variables, variable_types):
    var_types = []
    for variable in variables:
        if variable_types[variable] in ["b", "d", "o", "u"]:
            if variable_types[variable] in ["o", "u"]:
                var_types.append(variable_types[variable])
            else:
                var_types.append("u")
        elif variable_types[variable] in ["c"]:
            var_types.append("c")
        else:
            raise Exception(
                "Variable type {} for variable {} not a recognized type.".format(variable_types[variable], variable)
            )
    return "".join(var_types)


def binarize_discrete(data, covariates, variable_types):
    to_remove = []
    if variable_types:
        for variable in covariates:
            variable_type = variable_types[variable]
            if variable_type in ["d", "o", "u"]:
                dummies = get_dummies(data[variable])
                dummies.columns = [variable + str(col) for col in dummies.columns]
                dummies = dummies[dummies.columns[:-1]]
                covariates += list(dummies.columns)
                for var_name in dummies.columns:
                    variable_types[var_name] = "b"
                data[dummies.columns] = dummies
                to_remove.append(variable)
    for variable in to_remove:
        covariates.remove(variable)
        del data[variable]
    return data, covariates


def discrete_to_integer(discrete):
    encoder = LabelEncoder()
    discrete = encoder.fit_transform(discrete)
    return discrete, encoder



import networkx as nx
import numpy as np
import pandas as pd
from flaky import flaky
from pytest import approx
from sklearn.linear_model import LogisticRegression

from dowhy.gcm import StructuralCausalModel, auto, fit, intrinsic_causal_influence
from dowhy.gcm._noise import noise_samples_of_ancestors
from dowhy.gcm.graph import node_connected_subgraph_view
from dowhy.gcm.ml import create_hist_gradient_boost_classifier
from dowhy.gcm.uncertainty import estimate_entropy_of_probabilities, estimate_variance
from dowhy.gcm.util.general import apply_one_hot_encoding, fit_one_hot_encoders


@flaky(max_runs=3)
def test_intrinsic_causal_influence_variance_linear():
    causal_model = StructuralCausalModel(nx.DiGraph([("X0", "X1"), ("X1", "X2"), ("X2", "X3")]))

    X0 = np.random.normal(0, 1, 10000)
    X1 = X0 + np.random.normal(0, 0.001, 10000)
    X2 = X1 + np.random.normal(0, 2, 10000)
    X3 = X2 + np.random.normal(0, 1, 10000)
    training_data = pd.DataFrame({"X0": X0, "X1": X1, "X2": X2, "X3": X3})
    auto.assign_causal_mechanisms(causal_model, training_data, auto.AssignmentQuality.GOOD)

    fit(causal_model, training_data)

    iccs = intrinsic_causal_influence(causal_model, "X3", prediction_model="approx")
    assert iccs["X0"] == approx(1, abs=0.25)
    assert iccs["X1"] == approx(0, abs=0.05)
    assert iccs["X2"] == approx(4, abs=0.5)
    assert iccs["X3"] == approx(1, abs=0.25)
    assert np.sum([iccs[key] for key in iccs]) == approx(estimate_variance(X3), abs=0.5)

    iccs = intrinsic_causal_influence(causal_model, "X3", prediction_model="exact")
    assert iccs["X0"] == approx(1, abs=0.25)
    assert iccs["X1"] == approx(0, abs=0.05)
    assert iccs["X2"] == approx(4, abs=0.5)
    assert iccs["X3"] == approx(1, abs=0.25)
    assert np.sum([iccs[key] for key in iccs]) == approx(estimate_variance(X3), abs=0.5)


@flaky(max_runs=3)
def test_intrinsic_causal_influence_categorical():
    causal_model = StructuralCausalModel(nx.DiGraph([("X0", "X1"), ("X1", "X2"), ("X2", "X3")]))

    N1 = np.random.normal(0, 0.001, 10000)
    N2 = np.random.normal(0, 2, 10000)
    N3 = np.random.normal(0, 1, 10000)
    X0 = np.random.normal(0, 1, 10000)
    X1 = X0 + N1
    X2 = X1 + N2
    X3 = ((X2 + N3) >= 0).astype(str)
    training_data = pd.DataFrame({"X0": X0, "X1": X1, "X2": X2, "X3": X3})
    auto.assign_causal_mechanisms(causal_model, training_data, auto.AssignmentQuality.GOOD)

    fit(causal_model, training_data)

    log_mdl = LogisticRegression(max_iter=1000)
    log_mdl.fit(np.column_stack([X0, N1, N2, N3]), X3)

    expected_output_full_subset = -estimate_entropy_of_probabilities(
        log_mdl.predict_proba(np.column_stack([X0, N1, N2, N3]))
    )
    expected_output_empty_subset = -estimate_entropy_of_probabilities(np.array([[0.5, 0.5]]))

    iccs = intrinsic_causal_influence(causal_model, "X3", prediction_model="approx", num_samples_baseline=100)
    assert iccs["X0"] == approx(0.14, abs=0.05)
    assert iccs["X1"] == approx(0.01, abs=0.05)
    assert iccs["X2"] == approx(0.38, abs=0.05)
    assert iccs["X3"] == approx(0.12, abs=0.05)

    assert np.sum([iccs[key] for key in iccs]) == approx(
        expected_output_full_subset - expected_output_empty_subset, abs=0.05
    )

    iccs = intrinsic_causal_influence(causal_model, "X3", prediction_model="exact", num_samples_baseline=100)
    assert iccs["X0"] == approx(0.14, abs=0.05)
    assert iccs["X1"] == approx(0.01, abs=0.05)
    assert iccs["X2"] == approx(0.38, abs=0.05)
    assert iccs["X3"] == approx(0.12, abs=0.05)

    assert np.sum([iccs[key] for key in iccs]) == approx(-expected_output_empty_subset, abs=0.05)


@flaky(max_runs=3)
def test_intrinsic_causal_influence_categorical_2():
    causal_model = StructuralCausalModel(nx.DiGraph([("X0", "X1"), ("X1", "X2"), ("X2", "X3")]))
    N1 = np.random.normal(0, 0.001, 10000)
    N2 = np.random.normal(0, 2, 10000)
    X0 = np.random.normal(0, 1, 10000)
    X1 = X0 + N1
    X2 = X1 + N2

    X3 = []
    for i in range(10000):
        rand_val = np.random.choice(3, 1).squeeze()
        if rand_val == 0:
            X3.append(X2[i] + np.random.normal(-5, 0.5, 1))
        elif rand_val == 1:
            X3.append(X2[i] + np.random.normal(10, 0.5, 1))
        else:
            X3.append(X2[i] + np.random.normal(5, 0.5, 1))

    X3 = np.array(X3)
    X3 = (X3 > np.median(X3)).astype(str)
    training_data = pd.DataFrame({"X0": X0, "X1": X1, "X2": X2, "X3": X3.reshape(-1)})
    auto.assign_causal_mechanisms(causal_model, training_data, auto.AssignmentQuality.GOOD)

    fit(causal_model, training_data)

    data_samples, noise_samples = noise_samples_of_ancestors(
        StructuralCausalModel(node_connected_subgraph_view(causal_model.graph, "X3")), "X3", 100000
    )
    X = apply_one_hot_encoding(noise_samples.to_numpy(), fit_one_hot_encoders(noise_samples.to_numpy()))
    log_mdl = LogisticRegression()
    log_mdl.fit(X, data_samples["X3"].to_numpy())

    expected_output_full_subset = -estimate_entropy_of_probabilities(log_mdl.predict_proba(X))
    expected_output_empty_subset = -estimate_entropy_of_probabilities(np.array([[0.5, 0.5]]))

    iccs = intrinsic_causal_influence(causal_model, "X3", prediction_model="approx", num_samples_baseline=100)
    assert iccs["X0"] == approx(0.05, abs=0.05)
    assert iccs["X1"] == approx(0.03, abs=0.05)
    assert iccs["X2"] == approx(0.1, abs=0.05)
    assert iccs["X3"] == approx(0.5, abs=0.1)

    assert np.sum([iccs[key] for key in iccs]) == approx(
        expected_output_full_subset - expected_output_empty_subset, abs=0.05
    )

    iccs = intrinsic_causal_influence(causal_model, "X3", prediction_model="exact", num_samples_baseline=100)
    assert iccs["X0"] == approx(0.05, abs=0.05)
    assert iccs["X1"] == approx(0.03, abs=0.05)
    assert iccs["X2"] == approx(0.1, abs=0.05)
    assert iccs["X3"] == approx(0.5, abs=0.1)

    assert np.sum([iccs[key] for key in iccs]) == approx(-expected_output_empty_subset, abs=0.05)


@flaky(max_runs=3)
def test_given_only_categorical_data_when_estimate_icc_then_does_not_fail():
    causal_model = StructuralCausalModel(nx.DiGraph([("X0", "X1"), ("X1", "X2"), ("X2", "X3")]))
    X0 = np.random.normal(0, 1, 10000)
    X1 = X0 + np.random.normal(0, 0.001, 10000)
    X2 = X1 + np.random.normal(0, 2, 10000)
    X3 = ((X2 + np.random.normal(0, 1, 10000)) >= 0).astype(str)
    training_data = pd.DataFrame(
        {"X0": (X0 >= 0).astype(str), "X1": (X1 >= 0).astype(str), "X2": (X2 >= 0).astype(str), "X3": X3}
    )
    auto.assign_causal_mechanisms(causal_model, training_data, auto.AssignmentQuality.GOOD)

    fit(causal_model, training_data)

    data_samples, noise_samples = noise_samples_of_ancestors(
        StructuralCausalModel(node_connected_subgraph_view(causal_model.graph, "X3")), "X3", 100000
    )
    X = apply_one_hot_encoding(noise_samples.to_numpy(), fit_one_hot_encoders(noise_samples.to_numpy()))
    log_mdl = create_hist_gradient_boost_classifier()  # Due to the categorical features, this becomes non-linear.
    log_mdl.fit(X, data_samples["X3"].to_numpy())

    expected_output_full_subset = -estimate_entropy_of_probabilities(log_mdl.predict_probabilities(X))
    expected_output_empty_subset = -estimate_entropy_of_probabilities(np.array([[0.5, 0.5]]))

    iccs = intrinsic_causal_influence(causal_model, "X3", prediction_model="approx", num_samples_baseline=100)
    assert iccs["X0"] == approx(0.08, abs=0.03)
    assert iccs["X1"] == approx(0, abs=0.005)
    assert iccs["X2"] == approx(0.33, abs=0.05)
    assert iccs["X3"] == approx(0.27, abs=0.05)

    assert np.sum([iccs[key] for key in iccs]) == approx(
        expected_output_full_subset - expected_output_empty_subset, abs=0.05
    )

    iccs = intrinsic_causal_influence(causal_model, "X3", prediction_model="exact", num_samples_baseline=100)
    assert iccs["X0"] == approx(0.08, abs=0.03)
    assert iccs["X1"] == approx(0, abs=0.005)
    assert iccs["X2"] == approx(0.33, abs=0.05)
    assert iccs["X3"] == approx(0.27, abs=0.05)

    assert np.sum([iccs[key] for key in iccs]) == approx(-expected_output_empty_subset, abs=0.05)


def test_when_calling_intrinsic_causal_influence_then_the_shape_of_inputs_in_the_attribution_function_should_be_equal():
    causal_model = StructuralCausalModel(nx.DiGraph([("X0", "X1")]))
    X0 = np.random.normal(0, 1, 10000)
    X1 = X0 + np.random.normal(0, 0.001, 10000)
    training_data = pd.DataFrame({"X0": X0, "X1": X1})
    auto.assign_causal_mechanisms(causal_model, training_data, auto.AssignmentQuality.GOOD)

    fit(causal_model, training_data)

    def my_attr_func(X, Y):
        assert np.all(X.shape == Y.shape)
        return 0

    intrinsic_causal_influence(causal_model, "X1", attribution_func=my_attr_func)



"""Functions in this module should be considered experimental, meaning there might be breaking API changes in the
future.
"""
import numpy as np
from scipy.stats import entropy
from sklearn.neighbors import NearestNeighbors

from dowhy.gcm.constant import EPS
from dowhy.gcm.util.general import is_categorical, shape_into_2d


def auto_estimate_kl_divergence(X: np.ndarray, Y: np.ndarray) -> float:
    if is_categorical(X):
        return estimate_kl_divergence_categorical(X, Y)
    elif is_probability_matrix(X):
        return estimate_kl_divergence_of_probabilities(X, Y)
    else:
        return estimate_kl_divergence_continuous(X, Y)


def estimate_kl_divergence_continuous(X: np.ndarray, Y: np.ndarray) -> float:
    """Estimates KL-Divergence using k-nearest neighbours (Wang et al., 2009).

    Q. Wang, S. R. Kulkarni, and S. Verdú,
    "Divergence estimation for multidimensional densities via k-nearest-neighbor distances",
    IEEE Transactions on Information Theory, vol. 55, no. 5, pp. 2392-2405, May 2009.

    :param X: (N_1,D) Sample drawn from distribution P_X
    :param Y: (N_2,D) Sample drawn from distribution P_Y
    return: Estimated value of D(P_X||P_Y).
    """
    X, Y = shape_into_2d(X, Y)

    if X.shape[1] != Y.shape[1]:
        raise RuntimeError(
            "Samples from X and Y need to have the same dimension, but X has dimension %d and Y has "
            "dimension %d." % (X.shape[1], Y.shape[1])
        )

    n, m = X.shape[0], Y.shape[0]
    d = float(X.shape[1])

    k = int(np.sqrt(n))

    x_neighbourhood = NearestNeighbors(n_neighbors=k + 1).fit(X)
    y_neighbourhood = NearestNeighbors(n_neighbors=k).fit(Y)

    distances_x, _ = x_neighbourhood.kneighbors(X, n_neighbors=k + 1)
    distances_y, _ = y_neighbourhood.kneighbors(X, n_neighbors=k)

    rho = distances_x[:, -1]
    nu = distances_y[:, -1]

    result = np.sum((d / n) * np.log((nu + EPS) / (rho + EPS))) + np.log(m / (n - 1))

    if result < 0:
        result = 0

    return result


def estimate_kl_divergence_categorical(X: np.ndarray, Y: np.ndarray) -> float:
    X, Y = shape_into_2d(X, Y)

    if X.shape[1] != Y.shape[1]:
        raise RuntimeError(
            "Samples from X and Y need to have the same dimension, but X has dimension %d and Y has "
            "dimension %d." % (X.shape[1], Y.shape[1])
        )

    all_uniques = np.unique(np.vstack([X, Y]))

    p = np.array([(np.sum(X == i) + EPS) / (X.shape[0] + EPS) for i in all_uniques])
    q = np.array([(np.sum(Y == i) + EPS) / (Y.shape[0] + EPS) for i in all_uniques])

    return float(np.sum(p * np.log(p / q)))


def estimate_kl_divergence_of_probabilities(X: np.ndarray, Y: np.ndarray) -> float:
    """Estimates the Kullback-Leibler divergence between each pair of probability vectors (row wise) in X and Y
    separately and returns the mean over all results."""
    X, Y = shape_into_2d(X, Y)

    if X.shape[1] != Y.shape[1]:
        raise RuntimeError(
            "Samples from X and Y need to have the same dimension, but X has dimension %d and Y has "
            "dimension %d." % (X.shape[1], Y.shape[1])
        )

    return float(np.mean(entropy(X + EPS, Y + EPS, axis=1)))


def is_probability_matrix(X: np.ndarray) -> bool:
    if X.ndim == 1:
        return np.all(np.isclose(np.sum(abs(X), axis=0), 1))
    else:
        return np.all(np.isclose(np.sum(abs(X), axis=1), 1))



import itertools
import re

import numpy as np
import pandas as pd
import pytest
from sklearn.ensemble import GradientBoostingClassifier, GradientBoostingRegressor
from sklearn.linear_model import LogisticRegression, LogisticRegressionCV
from sklearn.model_selection import train_test_split
from sklearn.preprocessing import PolynomialFeatures

import dowhy
from dowhy import CausalModel, datasets

econml = pytest.importorskip("econml")


class TestEconMLEstimator:
    """Smoke tests for the integration with EconML estimators

    These tests only check that the ate estimation routine can be executed without errors.
    We don't check the accuracy of the ate estimates as we don't want to take dependencies on
    EconML estimators.
    """

    def test_backdoor_estimators(self):
        # Setup data
        data = datasets.linear_dataset(
            10,
            num_common_causes=4,
            num_samples=10000,
            num_instruments=2,
            num_effect_modifiers=2,
            num_treatments=1,
            treatment_is_binary=False,
        )
        df = data["df"]
        model = CausalModel(
            data=data["df"],
            treatment=data["treatment_name"],
            outcome=data["outcome_name"],
            effect_modifiers=data["effect_modifier_names"],
            graph=data["gml_graph"],
        )
        identified_estimand = model.identify_effect(proceed_when_unidentifiable=True)
        # Test LinearDML
        dml_estimate = model.estimate_effect(
            identified_estimand,
            method_name="backdoor.econml.dml.dml.LinearDML",
            control_value=0,
            treatment_value=1,
            target_units=lambda df: df["X0"] > 1,  # condition used for CATE
            method_params={
                "init_params": {
                    "model_y": GradientBoostingRegressor(),
                    "model_t": GradientBoostingRegressor(),
                    "featurizer": PolynomialFeatures(degree=1, include_bias=True),
                },
                "fit_params": {},
            },
        )
        # Test ContinuousTreatmentOrthoForest
        orthoforest_estimate = model.estimate_effect(
            identified_estimand,
            method_name="backdoor.econml.orf.DMLOrthoForest",
            target_units=lambda df: df["X0"] > 2,
            method_params={"init_params": {"n_trees": 10}, "fit_params": {}},
        )
        # Test LinearDRLearner
        data_binary = datasets.linear_dataset(
            10,
            num_common_causes=4,
            num_samples=10000,
            num_instruments=2,
            num_effect_modifiers=2,
            treatment_is_binary=True,
            outcome_is_binary=True,
        )
        model_binary = CausalModel(
            data=data_binary["df"],
            treatment=data_binary["treatment_name"],
            outcome=data_binary["outcome_name"],
            effect_modifiers=data["effect_modifier_names"],
            graph=data_binary["gml_graph"],
        )
        identified_estimand_binary = model_binary.identify_effect(proceed_when_unidentifiable=True)
        drlearner_estimate = model_binary.estimate_effect(
            identified_estimand_binary,
            method_name="backdoor.econml.dr.LinearDRLearner",
            target_units=lambda df: df["X0"] > 1,
            confidence_intervals=False,
            method_params={
                "init_params": {"model_propensity": LogisticRegressionCV(cv=3, solver="lbfgs", multi_class="auto")},
                "fit_params": {},
            },
        )

    def test_iv_estimators(self):
        keras = pytest.importorskip("keras")
        # Setup data
        data = datasets.linear_dataset(
            10,
            num_common_causes=4,
            num_samples=10000,
            num_instruments=2,
            num_effect_modifiers=2,
            num_treatments=1,
            treatment_is_binary=False,
        )
        df = data["df"]
        model = CausalModel(
            data=data["df"],
            treatment=data["treatment_name"],
            outcome=data["outcome_name"],
            effect_modifiers=data["effect_modifier_names"],
            graph=data["gml_graph"],
            identify_vars=True,
        )
        identified_estimand = model.identify_effect(proceed_when_unidentifiable=True)
        # Test DeepIV
        dims_zx = len(model._instruments) + len(model._effect_modifiers)
        dims_tx = len(model._treatment) + len(model._effect_modifiers)
        treatment_model = keras.Sequential(
            [
                keras.layers.Dense(128, activation="relu", input_shape=(dims_zx,)),  # sum of dims of Z and X
                keras.layers.Dropout(0.17),
                keras.layers.Dense(64, activation="relu"),
                keras.layers.Dropout(0.17),
                keras.layers.Dense(32, activation="relu"),
                keras.layers.Dropout(0.17),
            ]
        )
        response_model = keras.Sequential(
            [
                keras.layers.Dense(128, activation="relu", input_shape=(dims_tx,)),  # sum of dims of T and X
                keras.layers.Dropout(0.17),
                keras.layers.Dense(64, activation="relu"),
                keras.layers.Dropout(0.17),
                keras.layers.Dense(32, activation="relu"),
                keras.layers.Dropout(0.17),
                keras.layers.Dense(1),
            ]
        )
        deepiv_estimate = model.estimate_effect(
            identified_estimand,
            method_name="iv.econml.iv.nnet.DeepIV",
            target_units=lambda df: df["X0"] > -1,
            confidence_intervals=False,
            method_params={
                "init_params": {
                    "n_components": 10,  # Number of gaussians in the mixture density networks
                    # Treatment model,
                    "m": lambda z, x: treatment_model(keras.layers.concatenate([z, x])),
                    # Response model
                    "h": lambda t, x: response_model(keras.layers.concatenate([t, x])),
                    "n_samples": 1,  # Number of samples used to estimate the response
                    "first_stage_options": {"epochs": 25},
                    "second_stage_options": {"epochs": 25},
                },
                "fit_params": {},
            },
        )
        # Test IntentToTreatDRIV
        data = datasets.linear_dataset(
            10,
            num_common_causes=4,
            num_samples=10000,
            num_instruments=1,
            num_effect_modifiers=2,
            num_treatments=1,
            treatment_is_binary=True,
            num_discrete_instruments=1,
        )
        df = data["df"]
        model = CausalModel(
            data=data["df"],
            treatment=data["treatment_name"],
            outcome=data["outcome_name"],
            effect_modifiers=data["effect_modifier_names"],
            graph=data["gml_graph"],
        )
        identified_estimand = model.identify_effect(proceed_when_unidentifiable=True)
        driv_estimate = model.estimate_effect(
            identified_estimand,
            method_name="iv.econml.iv.dr.LinearIntentToTreatDRIV",
            target_units=lambda df: df["X0"] > 1,
            confidence_intervals=False,
            method_params={
                "init_params": {
                    "model_t_xwz": GradientBoostingClassifier(),
                    "model_y_xw": GradientBoostingRegressor(),
                    "flexible_model_effect": GradientBoostingRegressor(),
                    "featurizer": PolynomialFeatures(degree=1, include_bias=False),
                },
                "fit_params": {},
            },
        )

    def test_multivalue_treatment(self):
        n_points = 100000
        impact = {0: 0.0, 1: 2.0, 2: 1.0}
        df = pd.DataFrame(
            {
                "X": np.random.normal(size=n_points),
                "W": np.random.normal(size=n_points),
                "T": np.random.choice(np.array(list(impact.keys())), size=n_points),
            }
        )
        df["Y"] = df["W"] + df["T"].apply(lambda x: impact[x])

        train_data, test_data = train_test_split(df, train_size=0.9)

        causal_model = CausalModel(
            data=train_data,
            treatment="T",
            outcome="Y",
            common_causes="W",
            effect_modifiers="X",
        )
        identified_estimand = causal_model.identify_effect(proceed_when_unidentifiable=True)

        est_2 = causal_model.estimate_effect(
            identified_estimand,
            method_name="backdoor.econml.dml.dml.LinearDML",
            control_value=0,
            treatment_value=[1, 2],
            target_units="ate",  # condition used for CATE
            confidence_intervals=False,
            method_params={
                "init_params": {"discrete_treatment": True},
                "fit_params": {},
            },
        )

        est_test = est_2.estimator.effect_tt(test_data)

        est_error = (est_test - test_data["T"].apply(lambda x: impact[x]).values).abs().max()
        assert est_error < 0.03

    def test_empty_effect_modifiers(self):
        np.random.seed(101)
        data = dowhy.datasets.partially_linear_dataset(
            beta=10,
            num_common_causes=7,
            num_unobserved_common_causes=1,
            strength_unobserved_confounding=10,
            num_samples=1000,
            num_treatments=1,
            stddev_treatment_noise=10,
            stddev_outcome_noise=5,
        )

        # Observed data
        dropped_cols = ["W0"]
        user_data = data["df"].drop(dropped_cols, axis=1)
        # assumed graph
        user_graph = data["gml_graph"]
        for col in dropped_cols:
            user_graph = user_graph.replace('node[ id "{0}" label "{0}"]'.format(col), "")
            user_graph = re.sub('edge\[ source "{}" target "[vy][0]*"\]'.format(col), "", user_graph)

        model = CausalModel(
            data=user_data,
            treatment=data["treatment_name"],
            outcome=data["outcome_name"],
            graph=user_graph,
            test_significance=None,
        )

        model._graph.get_effect_modifiers(model._treatment, model._outcome)

        # Identify effect
        identified_estimand = model.identify_effect(proceed_when_unidentifiable=True)

        # Estimate effect
        model.estimate_effect(
            identified_estimand,
            method_name="backdoor.econml.dml.dml.LinearDML",
            method_params={
                "init_params": {
                    "model_y": GradientBoostingRegressor(),
                    "model_t": GradientBoostingRegressor(),
                    "linear_first_stages": False,
                },
                "fit_params": {
                    "cache_values": True,
                },
            },
        )



"""This module defines implementations of :class:`~dowhy.gcm.fcms.PredictionModel` used by the different
:class:`~dowhy.gcm.graph.FunctionalCausalModel` implementations, such as :class:`~dowhy.gcm.fcms.PostNonlinearModel` or
:class:`~dowhy.gcm.fcms.AdditiveNoiseModel`.
"""

from .classification import (
    SklearnClassificationModel,
    create_gaussian_process_classifier,
    create_hist_gradient_boost_classifier,
    create_logistic_regression_classifier,
    create_polynom_logistic_regression_classifier,
    create_random_forest_classifier,
)
from .regression import (
    SklearnRegressionModel,
    create_elastic_net_regressor,
    create_gaussian_process_regressor,
    create_hist_gradient_boost_regressor,
    create_lasso_lars_ic_regressor,
    create_lasso_regressor,
    create_linear_regressor,
    create_linear_regressor_with_given_parameters,
    create_polynom_regressor,
    create_random_forest_regressor,
    create_ridge_regressor,
    create_support_vector_regressor,
)

try:
    from dowhy.gcm.ml.autolguon import AutoGluonClassifier, AutoGluonRegressor
except ImportError:
    pass



import random

import numpy as np
import pytest
from flaky import flaky

from dowhy.gcm import generalised_cov_based
from tests.gcm.independence_test.test_kernel import _generate_categorical_data


@pytest.fixture
def preserve_random_generator_state():
    numpy_state = np.random.get_state()
    random_state = random.getstate()
    yield
    np.random.set_state(numpy_state)
    random.setstate(random_state)


@flaky(max_runs=5)
def test_given_conditional_independent_nonlinear_data_when_perform_generalised_cov_based_independence_test_then_returns_expected_result():
    z = np.random.randn(1000, 1)
    x = np.exp(z + np.random.rand(1000, 1))
    y = np.exp(z + np.random.rand(1000, 1))

    assert generalised_cov_based(x, y, z) > 0.05


@flaky(max_runs=5)
def test_given_dependent_nonlinear_data_when_perform_generalised_cov_based_independence_test_then_returns_expected_result():
    z = np.random.randn(1000, 1)
    w = np.random.randn(1000, 1)
    x = np.exp(z + np.random.rand(1000, 1))
    y = np.exp(z + np.random.rand(1000, 1))

    assert 0.05 > generalised_cov_based(x, y, w)


@flaky(max_runs=5)
def test_given_independent_categorical_Z_data_when_perform_generalised_cov_based_independence_test_then_returns_expected_result():
    x, y, z = _generate_categorical_data()

    assert generalised_cov_based(x, y, z) > 0.05


@flaky(max_runs=5)
def test_given_dependent_categorical_Z_data_when_perform_generalised_cov_based_independence_test_then_returns_expected_result():
    x, y, z = _generate_categorical_data()

    assert generalised_cov_based(x, y, z[np.random.choice(z.shape[0], z.shape[0], replace=False)]) < 0.05


@flaky(max_runs=5)
def test_given_unconditional_independent_nonlinear_data_when_perform_generalised_cov_based_independence_test_then_returns_expected_result():
    z = np.random.randn(1000, 1)
    w = np.random.randn(1000, 1)
    x = np.exp(z + np.random.rand(1000, 1))

    assert generalised_cov_based(x, w) > 0.05


@flaky(max_runs=5)
def test_given_unconditional_dependent_nonlinear_data_when_perform_generalised_cov_based_independence_test_then_returns_expected_result():
    z = np.random.randn(1000, 1)
    x = np.exp(z + np.random.rand(1000, 1))
    y = np.exp(z + np.random.rand(1000, 1))

    assert generalised_cov_based(x, y) < 0.05


def test_given_categorical_X_or_Y_when_perform_generalised_cov_based_independence_test_then_raise_error():
    with pytest.raises(ValueError):
        generalised_cov_based(np.random.choice(2, 1000, replace=True).astype(str), np.random.normal(0, 1, 1000))

    with pytest.raises(ValueError):
        generalised_cov_based(np.random.normal(0, 1, 1000), np.random.choice(2, 1000, replace=True).astype(str))



import numpy as np
from pytest import approx

from dowhy.gcm import MedianCDFQuantileScorer, MedianDeviationScorer


def test_given_simple_toy_data_when_using_MedianCDFQuantileScorer_then_returns_expected_scores():
    anomaly_scorer = MedianCDFQuantileScorer()
    anomaly_scorer.fit(np.array(range(0, 20)))
    assert anomaly_scorer.score(np.array([8, 17]))[0] == approx(1 - 17 / 20, abs=0.01)
    assert anomaly_scorer.score(np.array([0])) == approx(1 - 1 / 20, abs=0.01)

    anomaly_scorer.fit(np.array([4.0 for i in range(200)]))
    assert anomaly_scorer.score(np.array([4])) == approx(0, abs=0.01)


def test_given_simple_toy_data_when_using_MedianDeviationScorer_then_returns_expected_scores():
    anomaly_scorer = MedianDeviationScorer()
    anomaly_scorer.fit(np.array(range(0, 20)) / 10)
    assert anomaly_scorer.score(np.array([0.8, 1.7])).reshape(-1) == approx(np.array([0.2, 1]), abs=0.1)



"""This module provides functionality to estimate confidence intervals via bootstrapping the fitting and sampling.

Functions in this module should be considered experimental, meaning there might be breaking API changes in the future.
"""

from functools import partial
from typing import Any, Callable, Dict, Union

import numpy as np
import pandas as pd

from dowhy.gcm.cms import InvertibleStructuralCausalModel, ProbabilisticCausalModel, StructuralCausalModel
from dowhy.gcm.fitting_sampling import fit

# A convenience function when computing confidence intervals specifically for non-deterministic causal queries. This
# function evaluates the provided causal query multiple times to build a confidence interval based on the returned
# results.
# Note that this function does not re-fit the causal model(s) and only executes the provided query as it is. In order
# to re-refit the graphical causal model on random subsets of the data before executing the query, consider using the
# fit_and_compute function.
#
# **Example usage:**
#
#     >>> gcm.fit(causal_model, data)
#     >>> strength_medians, strength_intervals = gcm.confidence_intervals(
#     >>>     gcm.bootstrap_sampling(gcm.arrow_strength, causal_model, target_node='Y'))
#
# In this example, gcm.confidence_intervals is expecting a callable with non-deterministic outputs for building the
# confidence intervals. Since each causal query potentially expects a different set of parameters, we use 'partial'
# here to configure the function call. In this case,
# gcm.bootstrap_sampling(gcm.arrow_strength, causal_model, target_node='Y') would be equivalent to
# lambda : gcm.arrow_strength(causal_model, target_node='Y').
#
# In order to incorporate uncertainties coming from fitting the causal model(s), we can use
# gcm.fit_and_compute instead:
# >>>  strength_medians, strength_intervals = gcm.confidence_intervals(
# >>>        gcm.fit_and_compute(gcm.arrow_strength,
# >>>                                            causal_model,
# >>>                                            bootstrap_training_data=data,
# >>>                                            target_node='Y'))
# This would refit the provided causal_model on a subset of the data first before executing gcm.arrow_strength in each
# run.
bootstrap_sampling = partial


def fit_and_compute(
    f: Callable[
        [Union[ProbabilisticCausalModel, StructuralCausalModel, InvertibleStructuralCausalModel], Any],
        Dict[Any, Union[np.ndarray, float]],
    ],
    causal_model: Union[ProbabilisticCausalModel, StructuralCausalModel, InvertibleStructuralCausalModel],
    bootstrap_training_data: pd.DataFrame,
    bootstrap_data_subset_size_fraction: float = 0.75,
    *args,
    **kwargs,
):
    """A convenience function when computing confidence intervals specifically for causal queries. This function
    specifically bootstraps training *and* sampling.

    **Example usage:**

        >>> scores_median, scores_intervals = gcm.confidence_intervals(
        >>>     gcm.fit_and_compute(gcm.arrow_strength,
        >>>                         causal_model,
        >>>                         bootstrap_training_data=data,
        >>>                         target_node='Y'))

    :param f: The causal query to perform. A causal query is a function taking a graphical causal model as first
              parameter and an arbitrary number of remaining parameters. It must return a dictionary with
              attribution-like data.
    :param causal_model: A graphical causal model to perform the causal query on. It need not be fitted.
    :param bootstrap_training_data: The training data to use when fitting. A random subset from this data set is used
                                    in every iteration when calling fit.
    :param bootstrap_data_subset_size_fraction: The fraction defines the fractional size of the subset compared to
                                                the total training data.
    :param args: Args passed through verbatim to the causal queries.
    :param kwargs: Keyword args passed through verbatim to the causal queries.
    :return: A tuple containing (1) the median of causal query results and (2) the confidence intervals.
    """

    def snapshot():
        causal_model_copy = causal_model.clone()
        sampled_data = bootstrap_training_data.iloc[
            np.random.choice(
                bootstrap_training_data.shape[0],
                int(bootstrap_training_data.shape[0] * bootstrap_data_subset_size_fraction),
                replace=False,
            )
        ]
        fit(causal_model_copy, sampled_data)
        return f(causal_model_copy, *args, **kwargs)

    return snapshot



import numpy as np
import pandas as pd
import pytest
from pytest import mark
from sklearn.linear_model import LinearRegression

import dowhy.api
import dowhy.datasets


@mark.usefixtures("fixed_seed")
class TestPandasDoAPI(object):
    @mark.parametrize(
        ["N", "error_tolerance"],
        [
            (10000, 0.1),
        ],
    )
    def test_pandas_api_discrete_cause_continuous_confounder(self, N, error_tolerance):
        data = dowhy.datasets.linear_dataset(
            beta=10, num_common_causes=1, num_instruments=1, num_samples=N, treatment_is_binary=False
        )
        X0 = np.random.normal(size=N)
        v = (np.random.normal(size=N) + X0).astype(int)
        y = data["ate"] * v + X0 + np.random.normal()
        data["df"]["v"] = v
        data["df"]["X0"] = X0
        data["df"]["y"] = y
        df = data["df"].copy()

        variable_types = {"v": "d", "X0": "c", "y": "c"}
        outcome = "y"
        cause = "v"
        common_causes = "X0"
        method = "weighting"
        causal_df = df.causal.do(
            x=cause,
            variable_types=variable_types,
            outcome=outcome,
            method=method,
            common_causes=common_causes,
            proceed_when_unidentifiable=True,
        )

        ate = (causal_df[causal_df.v == 1].mean() - causal_df[causal_df.v == 0].mean())["y"]
        error = np.abs(ate - data["ate"])
        res = True if (error < data["ate"] * error_tolerance) else False
        print(
            "Error in ATE estimate = {0} with tolerance {1}%. Estimated={2},True={3}".format(
                error, error_tolerance * 100, ate, data["ate"]
            )
        )
        assert res

    @mark.parametrize(
        ["N", "error_tolerance"],
        [
            (10000, 0.1),
        ],
    )
    def test_pandas_api_discrete_cause_discrete_confounder(self, N, error_tolerance):
        data = dowhy.datasets.linear_dataset(
            beta=10, num_common_causes=1, num_instruments=1, num_samples=N, treatment_is_binary=False
        )
        X0 = np.random.normal(size=N).astype(int)
        v = (np.random.normal(size=N) + X0).astype(int)
        y = data["ate"] * v + X0 + np.random.normal()
        data["df"]["v"] = v
        data["df"]["X0"] = X0
        data["df"]["y"] = y
        df = data["df"].copy()

        variable_types = {"v": "d", "X0": "d", "y": "c"}
        outcome = "y"
        cause = "v"
        common_causes = "X0"
        method = "weighting"
        causal_df = df.causal.do(
            x=cause,
            variable_types=variable_types,
            outcome=outcome,
            method=method,
            common_causes=common_causes,
            proceed_when_unidentifiable=True,
        )

        ate = (causal_df[causal_df.v == 1].mean() - causal_df[causal_df.v == 0].mean())["y"]
        print("ate", ate)
        error = np.abs(ate - data["ate"])
        res = True if (error < data["ate"] * error_tolerance) else False
        print(
            "Error in ATE estimate = {0} with tolerance {1}%. Estimated={2},True={3}".format(
                error, error_tolerance * 100, ate, data["ate"]
            )
        )
        assert res

    @mark.parametrize(
        ["N", "error_tolerance"],
        [
            (10000, 0.1),
        ],
    )
    def test_pandas_api_continuous_cause_discrete_confounder(self, N, error_tolerance):
        data = dowhy.datasets.linear_dataset(
            beta=10, num_common_causes=1, num_instruments=1, num_samples=N, treatment_is_binary=False
        )
        X0 = np.random.normal(size=N).astype(int)
        v = np.random.normal(size=N) + X0
        y = data["ate"] * v + X0 + np.random.normal()
        data["df"]["v"] = v
        data["df"]["X0"] = X0
        data["df"]["y"] = y
        df = data["df"].copy()

        variable_types = {"v": "c", "X0": "d", "y": "c"}
        outcome = "y"
        cause = "v"
        common_causes = "X0"
        method = "weighting"
        causal_df = df.causal.do(
            x=cause,
            variable_types=variable_types,
            outcome=outcome,
            method=method,
            common_causes=common_causes,
            proceed_when_unidentifiable=True,
        )

        ate = LinearRegression().fit(causal_df[["v"]], causal_df["y"]).coef_[0]
        print("ate", ate)
        error = np.abs(ate - data["ate"])
        res = True if (error < data["ate"] * error_tolerance) else False
        print(
            "Error in ATE estimate = {0} with tolerance {1}%. Estimated={2},True={3}".format(
                error, error_tolerance * 100, ate, data["ate"]
            )
        )
        assert res

    @mark.parametrize(
        ["N", "error_tolerance"],
        [
            (10000, 0.1),
        ],
    )
    def test_pandas_api_continuous_cause_continuous_confounder(self, N, error_tolerance):
        data = dowhy.datasets.linear_dataset(
            beta=10, num_common_causes=1, num_instruments=1, num_samples=N, treatment_is_binary=False
        )
        X0 = np.random.normal(size=N)
        v = np.random.normal(size=N) + X0
        y = data["ate"] * v + X0 + np.random.normal()
        data["df"]["v"] = v
        data["df"]["X0"] = X0
        data["df"]["y"] = y
        df = data["df"].copy()

        variable_types = {"v": "c", "X0": "c", "y": "c"}
        outcome = "y"
        cause = "v"
        common_causes = "X0"
        method = "weighting"
        causal_df = df.causal.do(
            x=cause,
            variable_types=variable_types,
            outcome=outcome,
            method=method,
            common_causes=common_causes,
            proceed_when_unidentifiable=True,
        )

        ate = LinearRegression().fit(causal_df[["v"]], causal_df["y"]).coef_[0]
        print("ate", ate)
        error = np.abs(ate - data["ate"])
        res = True if (error < data["ate"] * error_tolerance) else False
        print(
            "Error in ATE estimate = {0} with tolerance {1}%. Estimated={2},True={3}".format(
                error, error_tolerance * 100, ate, data["ate"]
            )
        )
        assert res

    """
    In the following three tests, we have made use of the assert True at the end, but it is not
    a tautology due to the fact the function being tested has the ability to raise an exception
    when it belives that the behavior of the function is wrong.
    """

    @mark.parametrize(
        ["N", "variable_types"],
        [
            (10000, {"v0": "b", "y": "c", "W0": "c"}),
        ],
    )
    def test_pandas_api_with_full_specification_of_type(self, N, variable_types):
        data = dowhy.datasets.linear_dataset(
            beta=5, num_common_causes=1, num_instruments=0, num_samples=1000, treatment_is_binary=True
        )

        data["df"].causal.do(
            x="v0", variable_types=variable_types, outcome="y", proceed_when_unidentifiable=True, common_causes=["W0"]
        ).groupby("v0").mean()
        assert True

    @mark.parametrize(
        ["N", "variable_types"],
        [
            (10000, {"v0": "b", "W0": "c"}),
        ],
    )
    def test_pandas_api_with_partial_specification_of_type(self, N, variable_types):
        data = dowhy.datasets.linear_dataset(
            beta=5, num_common_causes=1, num_instruments=0, num_samples=1000, treatment_is_binary=True
        )

        data["df"].causal.do(
            x="v0", variable_types=variable_types, outcome="y", proceed_when_unidentifiable=True, common_causes=["W0"]
        ).groupby("v0").mean()
        assert True

    @mark.parametrize(
        ["N", "variable_types"],
        [
            (10000, {}),
        ],
    )
    def test_pandas_api_with_no_specification_of_type(self, N, variable_types):
        data = dowhy.datasets.linear_dataset(
            beta=5, num_common_causes=1, num_instruments=0, num_samples=1000, treatment_is_binary=True
        )

        data["df"].causal.do(
            x="v0", variable_types=variable_types, outcome="y", proceed_when_unidentifiable=True, common_causes=["W0"]
        ).groupby("v0").mean()
        assert True

    @mark.parametrize(
        ["N", "variable_types"],
        [
            (1, {"v0": "b", "W0": "c"}),
        ],
    )
    def test_pandas_api_with_dummy_data(self, N, variable_types):
        df = pd.DataFrame({"x": [0, 0.5, 1], "y": [1, 0.5, 0], "a": [0, 0.5, 0], "b": [0.25, 0, 0]})
        dd = df.causal.do(
            x=["x"],
            outcome="y",
            common_causes=["a", "b"],
            proceed_when_unidentifiable=True,
            variable_types=dict(x="c", y="c", a="c", b="c"),
        )
        print(dd)



"""This module provides functionality to answer what-if questions.

Functions in this module should be considered experimental, meaning there might be breaking API changes in the future.
"""

from typing import Any, Callable, Dict, Iterable, List, Optional, Union

import networkx as nx
import numpy as np
import pandas as pd

from dowhy.gcm._noise import compute_noise_from_data
from dowhy.gcm.cms import InvertibleStructuralCausalModel, ProbabilisticCausalModel, StructuralCausalModel
from dowhy.gcm.fcms import ClassifierFCM
from dowhy.gcm.fitting_sampling import draw_samples
from dowhy.gcm.graph import (
    DirectedGraph,
    get_ordered_predecessors,
    is_root_node,
    node_connected_subgraph_view,
    validate_causal_dag,
    validate_node_in_graph,
)


def interventional_samples(
    causal_model: ProbabilisticCausalModel,
    interventions: Dict[Any, Callable[[np.ndarray], Union[float, np.ndarray]]],
    observed_data: Optional[pd.DataFrame] = None,
    num_samples_to_draw: Optional[int] = None,
) -> pd.DataFrame:
    """Performs intervention on nodes in the causal graph.

    :param causal_model: The probabilistic causal model we perform this intervention on .
    :param interventions: Dictionary containing the interventions we want to perform, keyed by node name. An
                          intervention is a function that takes a value as input and returns another value.
                          For example, `{'X': lambda x: 2}` mimics the atomic intervention *do(X:=2)*.
                          A soft intervention can be formulated as `{'X': lambda x: 0.2 * x}`.
    :param observed_data: Optionally, data on which to perform interventions. If None are given, data is generated based
                          on the generative models.
    :param num_samples_to_draw: Sample size to draw from the interventional distribution.
    :return: Samples from the interventional distribution.
    """
    validate_causal_dag(causal_model.graph)
    for node in interventions:
        validate_node_in_graph(causal_model.graph, node)

    if observed_data is None and num_samples_to_draw is None:
        raise ValueError("Either observed_samples or num_samples_to_draw need to be set!")
    if observed_data is not None and num_samples_to_draw is not None:
        raise ValueError("Either observed_samples or num_samples_to_draw need to be set, not both!")

    if num_samples_to_draw is not None:
        observed_data = draw_samples(causal_model, num_samples_to_draw)

    return _interventional_samples(causal_model, observed_data, interventions)


def _interventional_samples(
    pcm: ProbabilisticCausalModel,
    observed_data: pd.DataFrame,
    interventions: Dict[Any, Callable[[np.ndarray], np.ndarray]],
) -> pd.DataFrame:
    samples = observed_data.copy()

    affected_nodes = _get_nodes_affected_by_intervention(pcm.graph, interventions.keys())
    sorted_nodes = nx.topological_sort(pcm.graph)

    # Simulating interventions by propagating the effects through the graph. For this, we iterate over the nodes based
    # on their topological order.
    for node in sorted_nodes:
        if node not in affected_nodes:
            continue

        if is_root_node(pcm.graph, node):
            node_data = samples[node].to_numpy()
        else:
            node_data = pcm.causal_mechanism(node).draw_samples(_parent_samples_of(node, pcm, samples))

        # After drawing samples of the node based on the data generation process, we apply the corresponding
        # intervention. The inputs of downstream nodes are therefore based on the outcome of the intervention in this
        # node.
        samples[node] = _evaluate_intervention(node, interventions, node_data.reshape(-1))

    return samples


def _get_nodes_affected_by_intervention(causal_graph: DirectedGraph, target_nodes: Iterable[Any]) -> List[Any]:
    result = []

    for node in nx.topological_sort(causal_graph):
        if node in target_nodes:
            result.append(node)
            continue

        for target_node in target_nodes:
            if target_node in nx.ancestors(causal_graph, source=node):
                result.append(node)
                break

    return result


def counterfactual_samples(
    causal_model: Union[StructuralCausalModel, InvertibleStructuralCausalModel],
    interventions: Dict[Any, Callable[[np.ndarray], Union[float, np.ndarray]]],
    observed_data: Optional[pd.DataFrame] = None,
    noise_data: Optional[pd.DataFrame] = None,
) -> pd.DataFrame:
    """Estimates counterfactual data for observed data if we were to perform specified interventions. This function
    implements the 3-step process for computing counterfactuals by Pearl (see https://ftp.cs.ucla.edu/pub/stat_ser/r485.pdf).

    :param causal_model: The (invertible) structural causal model we perform this intervention on. If noise_data is
                         None and observed_data is provided, this must be an invertible structural model, otherwise,
                         this can be either a structural causal model or an invertible one.
    :param interventions: Dictionary containing the interventions we want to perform keyed by node name. An
                          intervention is a function that takes a value as input and returns another value.
                          For example, `{'X': lambda x: 2}` mimics the atomic intervention *do(X:=2)*.
    :param observed_data: Factual data that we observe for the nodes in the causal graph.
    :param noise_data: Data of noise terms corresponding to nodes in the causal graph. If not provided,
                       these have to be estimated from observed data. Then we require causal models of nodes to be
                       invertible.
    :return: Estimated counterfactual data.
    """
    for node in interventions:
        validate_node_in_graph(causal_model.graph, node)

    validate_causal_dag(causal_model.graph)

    if observed_data is None and noise_data is None:
        raise ValueError("Either observed_data or noise_data need to be given!")
    if observed_data is not None and noise_data is not None:
        raise ValueError("Either observed_data or noise_data can be given, not both!")

    if noise_data is None and observed_data is not None:
        if not isinstance(causal_model, InvertibleStructuralCausalModel):
            raise ValueError(
                "Since no noise_data is given, this has to be estimated from the given "
                "observed_data. This can only be done with InvertibleStructuralCausalModel."
            )
        # Abduction: For invertible SCMs, we recover exact noise values from data.
        noise_data = compute_noise_from_data(causal_model, observed_data)

    # Action + Prediction: Propagate the intervention downstream using recovered noise values.
    return _counterfactual_samples(causal_model, interventions, noise_data)


def _counterfactual_samples(
    scm: StructuralCausalModel,
    interventions: Dict[Any, Callable[[np.ndarray], Union[float, np.ndarray]]],
    noise_data: pd.DataFrame,
) -> pd.DataFrame:
    topologically_sorted_nodes = list(nx.topological_sort(scm.graph))
    samples = pd.DataFrame(
        np.empty((noise_data.shape[0], len(topologically_sorted_nodes))), columns=topologically_sorted_nodes
    )

    for node in topologically_sorted_nodes:
        if is_root_node(scm.graph, node):
            node_data = noise_data[node].to_numpy()
        else:
            node_data = scm.causal_mechanism(node).evaluate(
                _parent_samples_of(node, scm, samples), noise_data[node].to_numpy()
            )

        samples[node] = _evaluate_intervention(node, interventions, node_data.reshape(-1))

    return samples


def _evaluate_intervention(
    node: Any, interventions: Dict[Any, Callable[[np.ndarray], np.ndarray]], pre_intervention_data: np.ndarray
) -> np.ndarray:
    # Check if we need to apply an intervention on the given node.
    if node in interventions:
        # Apply intervention function to the data of the node.
        post_intervention_data = np.array(list(map(interventions[node], pre_intervention_data)))

        # Check if the intervention function changes the shape of the data.
        if pre_intervention_data.shape != post_intervention_data.shape:
            raise RuntimeError(
                "Dimension of data corresponding to the node `%s` after intervention is different than before "
                "intervention." % node
            )

        return post_intervention_data
    else:
        return pre_intervention_data


def average_causal_effect(
    causal_model: ProbabilisticCausalModel,
    target_node: Any,
    interventions_alternative: Dict[Any, Callable[[np.ndarray], Union[float, np.ndarray]]],
    interventions_reference: Dict[Any, Callable[[np.ndarray], Union[float, np.ndarray]]],
    observed_data: Optional[pd.DataFrame] = None,
    num_samples_to_draw: Optional[int] = None,
) -> float:
    """Estimates the average causal effect (ACE) on the target of two different sets of interventions.
    The interventions can be specified through the parameters `interventions_alternative` and `interventions_reference`.
    For example, if the alternative intervention is do(T := 1) and the reference intervention
    is do(T := 0), then the average causal effect is given by ACE = E[Y | do(T := 1)] - E[Y | do(T := 0)]:
        >>> average_causal_effect(causal_model, 'Y', {'T': lambda _ : 1}, {'T': lambda _ : 0})

    We can also specify more complex interventions on multiple nodes:
        >>> average_causal_effect(causal_model,
        >>>                       'Y',
        >>>                       {'T': lambda _ : 1, 'X0': lambda x : x + 1},
        >>>                       {'T': lambda _ : 0, 'X0': lambda x : x * 2})
    In the above, we would estimate ACE = E[Y | do(T := 1), do(X0 := X0 + 1)] - E[Y | do(T := 0), do(X0 := X0 * 2)].

    Note: The target node can be a continuous real-valued variable or a categorical variable with at most two classes
    (i.e. binary).

    :param causal_model: The probabilistic causal model we perform this intervention on .
    :param target_node: Target node for which the ACE is estimated.
    :param interventions_alternative: Dictionary defining the interventions for the alternative values.
    :param interventions_reference: Dictionary defining the interventions for the reference values.
    :param observed_data: Factual data that we observe for the nodes in the causal graph. By default, new data
                          is sampled using the causal model. If observational data is available, providing them
                          might improve the accuracy by mitigating issues due to a misspecified graph and/or causal
                          models.
    :param num_samples_to_draw: Number of samples drawn from the causal model for estimating ACE if no observed data is
                                given.
    :return: The estimated average causal effect (ACE).
    """
    # For estimating the effect, we only need to consider the nodes that have a directed path to the target node, i.e.
    # all ancestors of the target.
    causal_model = ProbabilisticCausalModel(node_connected_subgraph_view(causal_model.graph, target_node))

    validate_causal_dag(causal_model.graph)
    for node in interventions_alternative:
        validate_node_in_graph(causal_model.graph, node)
    for node in interventions_reference:
        validate_node_in_graph(causal_model.graph, node)

    if observed_data is None and num_samples_to_draw is None:
        raise ValueError("Either observed_samples or num_samples_to_draw need to be set!")
    if observed_data is not None and num_samples_to_draw is not None:
        raise ValueError("Either observed_samples or num_samples_to_draw need to be set, not both!")

    if num_samples_to_draw is not None:
        observed_data = draw_samples(causal_model, num_samples_to_draw)

    samples_from_target_alt = _interventional_samples(causal_model, observed_data, interventions_alternative)[
        target_node
    ].to_numpy()
    samples_from_target_ref = _interventional_samples(causal_model, observed_data, interventions_reference)[
        target_node
    ].to_numpy()

    target_causal_model = causal_model.causal_mechanism(target_node)
    if isinstance(target_causal_model, ClassifierFCM):
        # The target node can be a continuous real-valued variable or a categorical variable with at most two classes
        # (i.e. binary).
        if observed_data[target_node].nunique() > 2:
            raise ValueError(
                "Cannot estimate average treatment effect of categorical data with more than 2 categories!"
            )

        class_names = target_causal_model.get_class_names(np.array([0, 1]))
        samples_from_target_alt[samples_from_target_alt == class_names[0]] = 0
        samples_from_target_alt[samples_from_target_alt == class_names[1]] = 1
        samples_from_target_ref[samples_from_target_ref == class_names[0]] = 0
        samples_from_target_ref[samples_from_target_ref == class_names[1]] = 1

    return np.mean(samples_from_target_alt) - np.mean(samples_from_target_ref)


def _parent_samples_of(node: Any, scm: ProbabilisticCausalModel, samples: pd.DataFrame) -> np.ndarray:
    return samples[get_ordered_predecessors(scm.graph, node)].to_numpy()



import logging
from typing import Optional, Union

import numpy as np
import pandas as pd
from joblib import Parallel, delayed
from tqdm.auto import tqdm

from dowhy.causal_estimator import CausalEstimate, CausalEstimator
from dowhy.causal_estimators.econml import Econml
from dowhy.causal_identifier import IdentifiedEstimand
from dowhy.causal_refuter import CausalRefutation, CausalRefuter, test_significance

logger = logging.getLogger(__name__)


class DataSubsetRefuter(CausalRefuter):
    """Refute an estimate by rerunning it on a random subset of the original data.

    Supports additional parameters that can be specified in the refute_estimate() method. For joblib-related parameters (n_jobs, verbose), please refer to the joblib documentation for more details (https://joblib.readthedocs.io/en/latest/generated/joblib.Parallel.html).

    :param subset_fraction: Fraction of the data to be used for re-estimation, which is ``DataSubsetRefuter.DEFAULT_SUBSET_FRACTION`` by default.
    :type  subset_fraction: float, optional

    :param num_simulations: The number of simulations to be run, which is ``CausalRefuter.DEFAULT_NUM_SIMULATIONS`` by default
    :type num_simulations: int, optional

    :param random_state: The seed value to be added if we wish to repeat the same random behavior. If we with to repeat the same behavior we push the same seed in the psuedo-random generator
    :type random_state: int, RandomState, optional

    :param n_jobs: The maximum number of concurrently running jobs. If -1 all CPUs are used. If 1 is given, no parallel computing code is used at all (this is the default).
    :type n_jobs: int, optional

    :param verbose: The verbosity level: if non zero, progress messages are printed. Above 50, the output is sent to stdout. The frequency of the messages increases with the verbosity level. If it more than 10, all iterations are reported. The default is 0.
    :type verbose: int, optional
    """

    # The default subset of the data to be used
    DEFAULT_SUBSET_FRACTION = 0.8

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._subset_fraction = kwargs.pop("subset_fraction", 0.8)
        self._num_simulations = kwargs.pop("num_simulations", CausalRefuter.DEFAULT_NUM_SIMULATIONS)
        self._random_state = kwargs.pop("random_state", None)

    def refute_estimate(self, show_progress_bar: bool = False):
        refute = refute_data_subset(
            data=self._data,
            target_estimand=self._target_estimand,
            estimate=self._estimate,
            subset_fraction=self._subset_fraction,
            num_simulations=self._num_simulations,
            random_state=self._random_state,
            show_progress_bar=show_progress_bar,
            n_jobs=self._n_jobs,
            verbose=self._verbose,
        )

        refute.add_refuter(self)
        return refute


def _refute_once(
    data: pd.DataFrame,
    target_estimand: IdentifiedEstimand,
    estimate: CausalEstimate,
    subset_fraction: float,
    random_state: Optional[Union[int, np.random.RandomState]],
):
    if random_state is None:
        new_data = data.sample(frac=subset_fraction)
    else:
        new_data = data.sample(frac=subset_fraction, random_state=random_state)

    new_estimator = estimate.estimator.get_new_estimator_object(target_estimand)
    new_estimator.fit(
        new_data,
        target_estimand.treatment_variable,
        target_estimand.outcome_variable,
        estimate.estimator._effect_modifier_names,
        **new_estimator._econml_fit_params if isinstance(new_estimator, Econml) else {},
    )
    new_effect = new_estimator.estimate_effect(
        control_value=estimate.control_value,
        treatment_value=estimate.treatment_value,
        target_units=estimate.estimator._target_units,
    )
    return new_effect.value


def refute_data_subset(
    data: pd.DataFrame,
    target_estimand: IdentifiedEstimand,
    estimate: CausalEstimate,
    subset_fraction: float = 0.8,
    num_simulations: int = 100,
    random_state: Optional[Union[int, np.random.RandomState]] = None,
    show_progress_bar: bool = False,
    n_jobs: int = 1,
    verbose: int = 0,
    **_,
) -> CausalRefutation:
    """Refute an estimate by rerunning it on a random subset of the original data.

    :param data: pd.DataFrame: Data to run the refutation
    :param target_estimand: IdentifiedEstimand: Identified estimand to run the refutation
    :param estimate: CausalEstimate: Estimate to run the refutation
    :param subset_fraction: Fraction of the data to be used for re-estimation, which is ``DataSubsetRefuter.DEFAULT_SUBSET_FRACTION`` by default.
    :param num_simulations: The number of simulations to be run, ``CausalRefuter.DEFAULT_NUM_SIMULATIONS`` by default
    :param random_state: The seed value to be added if we wish to repeat the same random behavior. For this purpose, we repeat the same seed in the psuedo-random generator.
    :param n_jobs: The maximum number of concurrently running jobs. If -1 all CPUs are used. If 1 is given, no parallel computing code is used at all (this is the default).
    :param verbose: The verbosity level: if non zero, progress messages are printed. Above 50, the output is sent to stdout. The frequency of the messages increases with the verbosity level. If it more than 10, all iterations are reported. The default is 0.
    """

    logger.info(
        "Refutation over {} simulated datasets of size {} each".format(
            subset_fraction, subset_fraction * len(data.index)
        )
    )

    # Run refutation in parallel
    sample_estimates = Parallel(n_jobs=n_jobs, verbose=verbose)(
        delayed(_refute_once)(data, target_estimand, estimate, subset_fraction, random_state)
        for _ in tqdm(
            range(num_simulations),
            colour=CausalRefuter.PROGRESS_BAR_COLOR,
            disable=not show_progress_bar,
            desc="Refuting Estimates: ",
        )
    )
    sample_estimates = np.array(sample_estimates)

    refute = CausalRefutation(estimate.value, np.mean(sample_estimates), refutation_type="Refute: Use a subset of data")

    # We want to see if the estimate falls in the same distribution as the one generated by the refuter
    # Ideally that should be the case as choosing a subset should not have a significant effect on the ability
    # of the treatment to affect the outcome
    refute.add_significance_test_results(test_significance(estimate, sample_estimates))

    return refute



from typing import Any, List, Optional, Union

import numpy as np
import pandas as pd
import statsmodels.api as sm

from dowhy.causal_estimator import CausalEstimate, CausalEstimator, IdentifiedEstimand


class RegressionEstimator(CausalEstimator):
    """Compute effect of treatment using some regression function.

    Fits a regression model for estimating the outcome using treatment(s) and
    confounders.

    Base class for all regression models, inherited by
    LinearRegressionEstimator and GeneralizedLinearModelEstimator.

    """

    def __init__(
        self,
        identified_estimand: IdentifiedEstimand,
        test_significance: bool = False,
        evaluate_effect_strength: bool = False,
        confidence_intervals: bool = False,
        num_null_simulations: int = CausalEstimator.DEFAULT_NUMBER_OF_SIMULATIONS_STAT_TEST,
        num_simulations: int = CausalEstimator.DEFAULT_NUMBER_OF_SIMULATIONS_CI,
        sample_size_fraction: int = CausalEstimator.DEFAULT_SAMPLE_SIZE_FRACTION,
        confidence_level: float = CausalEstimator.DEFAULT_CONFIDENCE_LEVEL,
        need_conditional_estimates: Union[bool, str] = "auto",
        num_quantiles_to_discretize_cont_cols: int = CausalEstimator.NUM_QUANTILES_TO_DISCRETIZE_CONT_COLS,
        **kwargs,
    ):
        """
        :param identified_estimand: probability expression
            representing the target identified estimand to estimate.
        :param test_significance: Binary flag or a string indicating whether to test significance and by which method. All estimators support test_significance="bootstrap" that estimates a p-value for the obtained estimate using the bootstrap method. Individual estimators can override this to support custom testing methods. The bootstrap method supports an optional parameter, num_null_simulations. If False, no testing is done. If True, significance of the estimate is tested using the custom method if available, otherwise by bootstrap.
        :param evaluate_effect_strength: (Experimental) whether to evaluate the strength of effect
        :param confidence_intervals: Binary flag or a string indicating whether the confidence intervals should be computed and which method should be used. All methods support estimation of confidence intervals using the bootstrap method by using the parameter confidence_intervals="bootstrap". The bootstrap method takes in two arguments (num_simulations and sample_size_fraction) that can be optionally specified in the params dictionary. Estimators may also override this to implement their own confidence interval method. If this parameter is False, no confidence intervals are computed. If True, confidence intervals are computed by the estimator's specific method if available, otherwise through bootstrap
        :param num_null_simulations: The number of simulations for testing the
            statistical significance of the estimator
        :param num_simulations: The number of simulations for finding the
            confidence interval (and/or standard error) for a estimate
        :param sample_size_fraction: The size of the sample for the bootstrap
            estimator
        :param confidence_level: The confidence level of the confidence
            interval estimate
        :param need_conditional_estimates: Boolean flag indicating whether
            conditional estimates should be computed. Defaults to True if
            there are effect modifiers in the graph
        :param num_quantiles_to_discretize_cont_cols: The number of quantiles
            into which a numeric effect modifier is split, to enable
            estimation of conditional treatment effect over it.
        :param kwargs: (optional) Additional estimator-specific parameters
        """
        super().__init__(
            identified_estimand=identified_estimand,
            test_significance=test_significance,
            evaluate_effect_strength=evaluate_effect_strength,
            confidence_intervals=confidence_intervals,
            num_null_simulations=num_null_simulations,
            num_simulations=num_simulations,
            sample_size_fraction=sample_size_fraction,
            confidence_level=confidence_level,
            need_conditional_estimates=need_conditional_estimates,
            num_quantiles_to_discretize_cont_cols=num_quantiles_to_discretize_cont_cols,
            **kwargs,
        )

        self.model = None

    def fit(
        self,
        data: pd.DataFrame,
        treatment_name: str,
        outcome_name: str,
        effect_modifier_names: Optional[List[str]] = None,
    ):
        """
        Fits the estimator with data for effect estimation
        :param data: data frame containing the data
        :param treatment: name of the treatment variable
        :param outcome: name of the outcome variable
        :param effect_modifiers: Variables on which to compute separate
                    effects, or return a heterogeneous effect function. Not all
                    methods support this currently.
        """
        self._set_data(data, treatment_name, outcome_name)
        self._set_effect_modifiers(effect_modifier_names)

        self.logger.debug("Back-door variables used:" + ",".join(self._target_estimand.get_backdoor_variables()))
        self._observed_common_causes_names = self._target_estimand.get_backdoor_variables()
        if len(self._observed_common_causes_names) > 0:
            self._observed_common_causes = self._data[self._observed_common_causes_names]
            self._observed_common_causes = pd.get_dummies(self._observed_common_causes, drop_first=True)
        else:
            self._observed_common_causes = None

        self.symbolic_estimator = self.construct_symbolic_estimator(self._target_estimand)
        self.logger.info(self.symbolic_estimator)

        # The model is always built on the entire data
        _, self.model = self._build_model()
        coefficients = self.model.params[1:]  # first coefficient is the intercept
        self.logger.debug("Coefficients of the fitted model: " + ",".join(map(str, coefficients)))
        self.logger.debug(self.model.summary())

        return self

    def estimate_effect(
        self,
        data: pd.DataFrame = None,
        treatment_value: Any = 1,
        control_value: Any = 0,
        target_units=None,
        need_conditional_estimates=None,
        **_,
    ):
        if data is None:
            data = self._data
        self._target_units = target_units
        self._treatment_value = treatment_value
        self._control_value = control_value
        # TODO make treatment_value and control value also as local parameters
        # All treatments are set to the same constant value
        effect_estimate = self._do(treatment_value, data) - self._do(control_value, data)
        conditional_effect_estimates = None
        if need_conditional_estimates:
            conditional_effect_estimates = self._estimate_conditional_effects(
                self._estimate_effect_fn, effect_modifier_names=self._effect_modifier_names
            )
        intercept_parameter = self.model.params[0]
        estimate = CausalEstimate(
            estimate=effect_estimate,
            control_value=control_value,
            treatment_value=treatment_value,
            conditional_estimates=conditional_effect_estimates,
            target_estimand=self._target_estimand,
            realized_estimand_expr=self.symbolic_estimator,
            intercept=intercept_parameter,
        )

        estimate.add_estimator(self)
        return estimate

    def _estimate_effect_fn(self, data_df):
        est = self.estimate_effect(data=data_df, need_conditional_estimates=False)
        return est.value

    def _build_features(self, treatment_values=None, data_df=None):
        # Using all data by default
        if data_df is None:
            data_df = self._data
            treatment_vals = pd.get_dummies(self._treatment, drop_first=True)
            observed_common_causes_vals = self._observed_common_causes
            effect_modifiers_vals = self._effect_modifiers
        else:
            treatment_vals = pd.get_dummies(data_df[self._treatment_name], drop_first=True)
            if len(self._observed_common_causes_names) > 0:
                observed_common_causes_vals = data_df[self._observed_common_causes_names]
                observed_common_causes_vals = pd.get_dummies(observed_common_causes_vals, drop_first=True)
            if self._effect_modifier_names:
                effect_modifiers_vals = data_df[self._effect_modifier_names]
                effect_modifiers_vals = pd.get_dummies(effect_modifiers_vals, drop_first=True)
        # Fixing treatment value to the specified value, if provided
        if treatment_values is not None:
            treatment_vals = treatment_values
        if type(treatment_vals) is not np.ndarray:
            treatment_vals = treatment_vals.to_numpy()
        # treatment_vals and data_df should have same number of rows
        if treatment_vals.shape[0] != data_df.shape[0]:
            raise ValueError("Provided treatment values and dataframe should have the same length.")
        # Bulding the feature matrix
        n_treatment_cols = 1 if len(treatment_vals.shape) == 1 else treatment_vals.shape[1]
        n_samples = treatment_vals.shape[0]
        treatment_2d = treatment_vals.reshape((n_samples, n_treatment_cols))
        if len(self._observed_common_causes_names) > 0:
            features = np.concatenate((treatment_2d, observed_common_causes_vals), axis=1)
        else:
            features = treatment_2d
        if self._effect_modifier_names:
            for i in range(treatment_2d.shape[1]):
                curr_treatment = treatment_2d[:, i]
                new_features = curr_treatment[:, np.newaxis] * effect_modifiers_vals.to_numpy()
                features = np.concatenate((features, new_features), axis=1)
        features = features.astype(
            float, copy=False
        )  # converting to float in case of binary treatment and no other variables
        features = sm.add_constant(features, has_constant="add")  # to add an intercept term
        return features

    def _do(self, treatment_val, data_df=None):
        if data_df is None:
            data_df = self._data
        if not self.model:
            # The model is always built on the entire data
            _, self.model = self._build_model()
        # Replacing treatment values by given x
        # First, create interventional tensor in original space
        interventional_treatment_values = np.full((data_df.shape[0], len(self._treatment_name)), treatment_val)
        # Then, use pandas to ensure that the dummies are assigned correctly for a categorical treatment
        interventional_treatment_2d = pd.concat(
            [
                self._treatment.copy(),
                pd.DataFrame(data=interventional_treatment_values, columns=self._treatment.columns),
            ],
            axis=0,
        ).astype(self._treatment.dtypes, copy=False)
        interventional_treatment_2d = pd.get_dummies(interventional_treatment_2d, drop_first=True)
        interventional_treatment_2d = interventional_treatment_2d[self._treatment.shape[0] :]

        new_features = self._build_features(treatment_values=interventional_treatment_2d, data_df=data_df)
        interventional_outcomes = self.predict_fn(self.model, new_features)
        return interventional_outcomes.mean()



"""Functions and classes in this module should be considered experimental, meaning there might be breaking API changes
in the future.
"""

from typing import List

import numpy as np
import sklearn
from packaging import version
from sklearn.pipeline import make_pipeline
from sklearn.preprocessing import PolynomialFeatures

if version.parse(sklearn.__version__) < version.parse("1.0"):
    from sklearn.experimental import enable_hist_gradient_boosting  # noqa

from sklearn.ensemble import (
    AdaBoostClassifier,
    ExtraTreesClassifier,
    HistGradientBoostingClassifier,
    RandomForestClassifier,
)
from sklearn.gaussian_process import GaussianProcessClassifier
from sklearn.linear_model import LogisticRegression
from sklearn.naive_bayes import GaussianNB
from sklearn.neighbors import KNeighborsClassifier
from sklearn.svm import SVC

from dowhy.gcm.fcms import ClassificationModel
from dowhy.gcm.ml.regression import SklearnRegressionModel
from dowhy.gcm.util.general import apply_one_hot_encoding, shape_into_2d


class SklearnClassificationModel(SklearnRegressionModel, ClassificationModel):
    def predict_probabilities(self, X: np.array) -> np.ndarray:
        return shape_into_2d(self._sklearn_mdl.predict_proba(apply_one_hot_encoding(X, self._one_hot_encoders)))

    @property
    def classes(self) -> List[str]:
        return self._sklearn_mdl.classes_

    def clone(self):
        return SklearnClassificationModel(sklearn_mdl=sklearn.clone(self._sklearn_mdl))


def create_random_forest_classifier(**kwargs) -> SklearnClassificationModel:
    return SklearnClassificationModel(RandomForestClassifier(**kwargs))


def create_gaussian_process_classifier(**kwargs) -> SklearnClassificationModel:
    return SklearnClassificationModel(GaussianProcessClassifier(**kwargs))


def create_hist_gradient_boost_classifier(**kwargs) -> SklearnClassificationModel:
    return SklearnClassificationModel(HistGradientBoostingClassifier(**kwargs))


def create_logistic_regression_classifier(**kwargs) -> SklearnClassificationModel:
    return SklearnClassificationModel(LogisticRegression(**kwargs))


def create_extra_trees_classifier(**kwargs) -> SklearnClassificationModel:
    return SklearnClassificationModel(ExtraTreesClassifier(**kwargs))


def create_ada_boost_classifier(**kwargs) -> SklearnClassificationModel:
    return SklearnClassificationModel(AdaBoostClassifier(**kwargs))


def create_support_vector_classifier(**kwargs) -> SklearnClassificationModel:
    return SklearnClassificationModel(SVC(**kwargs))


def create_knn_classifier(**kwargs) -> SklearnClassificationModel:
    return SklearnClassificationModel(KNeighborsClassifier(**kwargs))


def create_gaussian_nb_classifier(**kwargs) -> SklearnClassificationModel:
    return SklearnClassificationModel(GaussianNB(**kwargs))


def create_polynom_logistic_regression_classifier(
    degree: int = 3, **kwargs_logistic_regression
) -> SklearnClassificationModel:
    return SklearnClassificationModel(
        make_pipeline(
            PolynomialFeatures(degree=degree, include_bias=False), LogisticRegression(**kwargs_logistic_regression)
        )
    )
